/*
 *****************************************************************************
 * audio_rc.c
 Title: audio signal acquisition from radio control transmitter
 Authors:
  Joel Lienard
  Lionel Cailler
  greg
 
 Purpose:  Audio signal acquisition from radio control transmitter.
  This code is based on original  Joel Lienard source code modified to support a
  portable library to a wide list of operating systems : http://www.portaudio.com
 
 It's free, it works, enjoy.  This  is an open source project, so if you  don't
 like the way something works, help us fix it!
 
 
 * CHANGE HISTORY:
   06feb2005 Update of initialization phase : add a check of an available audio input device before using it.
             permits automatic detection of computer configurations without audio input device.
   03nov2003 add recompile with port_audio_v18_1 and text display improvements.
   30dec2003 correction made in audio_rc.c by greg : correction made in audio_rc.c 
    1)   while testing audio_rc.c independently of crrcsim, i saw that it generated Pa_Audio errors.
        this was corrected by moving the call to Pa_Initialize() before the call to the first Pa_Audio
        function, Pa_GetDefaultInputDeviceID().  however, the interface was still returnign zero values.
    3)  i then added a call to Pa_Sleep(1) before the loop calling get_data() in
        get_data_from_audio_interface().  now the audio interface was active and responding well.
   06aug2004 improvement of ppm signal decoding. Initial version get some acquisition values lost.
        Signal decoding and then, plane control, is more precise.
 
 *
 *****************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include "portaudio.h"
#include "audio_rc.h"

#include <stdio.h>
#include <stdlib.h>

#define SAMPLE_RATE  (44100)
#define NUM_SAMPLES   (3000)
#define NUM_CHANNELS    (2)

typedef float SAMPLE;

typedef struct
{
  int     frameIndex;  /* Index into sample array. */
  int     frameSize;
  int     SamplesCounter;
  int     LastSamplesCounter;
  SAMPLE  *recordedSamples;
} paAudioData;

paAudioData AudioData;
int synchro_index = 0;
PortAudioStream *stream;
int inputDevice;

/**
 * return sample number i of  synchronised audio signal
 */
float get_audio_signal(int i)
{
  float *sig;
  int j;
  sig = AudioData.recordedSamples;
  j = synchro_index + i;
  if ( j >=  AudioData.frameSize)
  {
    j -=  AudioData.frameSize;
  }
  return sig[j];                     
}

// --------------------------------------------------------------------
/* This routine will be called by the PortAudio engine when audio is needed.
** It may be called at interrupt level on some machines so don't do anything
** that could mess up the system like calling malloc() or free().
*/

/* purpose of this routine is to update a rotating buffer
 *  index of buffer is AudioData.frameIndex
 *  buffer size is AudioData.frameSize
 *  number of samples stored since audio_rc_open() called
 */
static int recordCallback( void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           PaTimestamp outTime, void *userData )
{
  paAudioData *data = (paAudioData*)userData;
  SAMPLE *rptr = (float*)inputBuffer;
  SAMPLE *wptr;
  unsigned long i;
  unsigned long framesLeft = data->frameSize - data->frameIndex;

  /* Prevent unused variable warnings. */
  outputBuffer = NULL;
  outTime = 0;

  data->SamplesCounter += framesPerBuffer;

  wptr = &data->recordedSamples[data->frameIndex];
  if( framesPerBuffer <= framesLeft )
  {
    for( i=0; i<framesPerBuffer; i++ )
    {
      *wptr++ = *rptr++;
      rptr++;
    }  // assume NUM_CHANNELS = 2
  }
  else
  {
    for( i=0; i<framesLeft; i++ )
    {
      *wptr++=*rptr++;
      rptr++;
    } // assume NUM_CHANNELS = 2

    wptr = &data->recordedSamples[0];
    for( i=0; i<framesPerBuffer-framesLeft; i++ )
    {
      *wptr++=*rptr++;
      rptr++;
    }  // assume NUM_CHANNELS = 2
  }

  data->frameIndex=(data->frameIndex+framesPerBuffer)%data->frameSize;

  return 0;
}


// --------------------------------------------------------------------
static int get_data(float *values,int *nvalues)
{
  float x, px, max, min, moy;
  int nval = AudioData.frameSize;
  int i, j, chanel=0;
  float time, dt;
  float *sig=AudioData.recordedSamples;

  max= -100000;
  min=  100000;
  for(i=0;i<nval;i++)
  {
    x = sig[i];
    if (x>max)
      max=x;
    if (x<min)
      min=x;
  }
  moy = (max+min)/2;
  x=0;
  chanel=-1;
  time = 0;
  j=(AudioData.frameIndex+1024)%AudioData.frameSize;
  for(i=0;i<nval;i++)
  {
    time++;
    if (time >100)
    {                   // Synchro detection
      if (chanel < 0)
        chanel=0; //start
      else if (chanel > 0)
        break; //end
    }
    px=x;
    x = sig[j++];
    if (j==AudioData.frameSize)
    {
      j=0;
    }
    if((x>moy) && (px<moy) )
    {
      dt = (moy-px)/(x-px);
      if (chanel>=0)
      {
        if(chanel==0 && i<nval/2)
        {  
          synchro_index = j;  //for OSCILLO  synchro
        }
        if(chanel>0)
        {
          values[chanel-1]=(time+dt)/44.1-1.5;
        }
        chanel ++;
        if(chanel>10)
          break;
      }
      time = -dt;
    }
  }
  *nvalues=chanel-1;

  return 1;
}

// --------------------------------------------------------------------
int get_data_from_audio_interface(float *values)
{
  static float vals[11];
  static int nvals=0;
  int i,n;

  /*
   * there is no need to update values more than 50 times a second
   * then lest check here every (SAMPLE_RATE/50) recorded samples
   */

  Pa_Sleep (1); // not sure this is usefull

  n=AudioData.SamplesCounter;
  if (n-AudioData.LastSamplesCounter>SAMPLE_RATE/50)
  {

    AudioData.LastSamplesCounter = n;                         // lets update values
    if(get_data(vals,&n))
    {
      for(i=0;i<11;i++)
        values[i]=vals[i];
      nvals = n;
    }
  }
  else
    /*
     * LastSamplesCounter & SamplesCounter saturation management
     */
    if (n<AudioData.LastSamplesCounter)
      AudioData.LastSamplesCounter = n;

  return nvals;
}

// --------------------------------------------------------------------
int audio_rc_open(T_TX_InterfaceAudio *iface)
{
  int i,numDevices;
  PaError    err;
  PaDeviceInfo pdi;

  err = Pa_Initialize();
  if( err != paNoError )
    goto error;

/* List all available audio input and output devices */
 
  numDevices = Pa_CountDevices();
  printf(    "Number of total audio input/output devices : %d.\n", numDevices );
  for( i=0; i<numDevices; i++ )
  {
    pdi = *Pa_GetDeviceInfo( i );
    printf("Device(%d) name  \t: %s.\n", i, pdi.name);
    printf("\tMax Inputs   = %d\n", pdi.maxInputChannels  );
    printf("\tMax Outputs  = %d\n", pdi.maxOutputChannels  );
  }

/* display default input devices indentifier */
  if (iface->sc_num > numDevices)
  {
    fprintf(stderr, "Soundcard specified in configuration file is invalid (%d)\n", iface->sc_num);
    inputDevice = Pa_GetDefaultInputDeviceID();
  }
  else
  {
    inputDevice = iface->sc_num;
  }
  pdi = *Pa_GetDeviceInfo( inputDevice );
  printf("Using input device with ID %d (%s)\n", inputDevice, pdi.name );

/*  check configuration */

  if (pdi.maxInputChannels<NUM_CHANNELS)
  {
    printf("no audio input device available, please configure your computer to support an input audio device.\n");
    Pa_Terminate();
    return (-1);
  }

/*  configure device */

  AudioData.frameSize = NUM_SAMPLES; /* Record for a few samples. */
  AudioData.frameIndex = 0;
  AudioData.LastSamplesCounter=0;
  AudioData.SamplesCounter=0;

  AudioData.recordedSamples = (SAMPLE *) malloc( AudioData.frameSize * sizeof(SAMPLE) );
  if( AudioData.recordedSamples == NULL )
  {
    printf("Could not allocate record array.\n");
    exit(1);
  }
  for( i=0; i<AudioData.frameSize; i++ )
    AudioData.recordedSamples[i] = 0;


  /* Record some audio. -------------------------------------------- */
  err = Pa_OpenStream(
          &stream,
          inputDevice,
          NUM_CHANNELS,               /* NUM_CHANNELS : stereo input */
          paFloat32,
          NULL,
          paNoDevice,
          0,
          paFloat32,
          NULL,
          SAMPLE_RATE,
          1024,            /* frames per buffer */
          0,               /* number of buffers, if zero then use default minimum */
          0, //paDitherOff,    /* flags */
          recordCallback,
          &AudioData );
  if( err != paNoError )
    goto error;

  err = Pa_StartStream( stream );
  if( err != paNoError )
    goto error;

  // save the interface number for the configuration
  iface->sc_num = inputDevice;
  
  return 0;

error:
  Pa_Terminate();
  fprintf( stderr, "An error occured while using the portaudio stream in audio_rc_open\n" );
  fprintf( stderr, "Error number: %d\n", err );
  fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );

  return -1;
}

// --------------------------------------------------------------------
int audio_rc_close()
{
  PaError    err;

  if (stream!=NULL)
  {
    err = Pa_CloseStream( stream );
    if( err != paNoError )
      goto error;
  }
  if (AudioData.recordedSamples!=NULL)
    free( AudioData.recordedSamples );

  Pa_Terminate();
  
  printf("audio device closed\n");
  
  return 0;

error:
  Pa_Terminate();
  fprintf( stderr, "An error occured while using the portaudio stream in audio_rc_close\n" );
  fprintf( stderr, "Error number: %d\n", err );
  fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );

  return -1;
}

