//----------------------------------------------------------------------------------//
//                                                                                  //
//  Created on November 09, 2006                                                    //
//                                                                                  // 
//  "Copyright (c) 2006 The Regents of the University of California.                //
//  All rights reserved.                                                            //
//                                                                                  //
//  Permission to use, copy, modify, and distribute this software and its           //
//  documentation for any purpose, without fee, and without written agreement is    //
//  hereby granted, provided that the above copyright notice, the following         //
//  two paragraphs and the author appear in all copies of this software.            //
//                                                                                  //
//  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR       //
//  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT     //
//  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF    //
//  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                  //
//                                                                                  //
//  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,             //
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY        //
//  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS       //
//  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO      //
//  PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."         //
//                                                                                  //
//----------------------------------------------------------------------------------//


//----------------------------------------------------------------------------------
//
// chardevice.h
// 
//    Cross-platform one-to-one communication over serial, TCP (client and server),
//    and UDP (client and server) through a simple read/write interface that is
//    robust to disconnection. Also a "wrapper" char device that allows any one of
//    the above to be used transparently, and a "buffered character device" that
//    serves as a base class for message-passing architectures that receive a single
//    type of message that is no more than 256 characters in length (transparently
//    over any one of the supported interfaces).
//
// authors
//
//    Todd R. Templeton <ttemplet@eecs.berkeley.edu>
//
//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------
//
// Status
//
// Linux, g++: Tested and working
// Linux PPC, g++: TCP/UDP tested and working, serial untested
// Windows XP, MS Visual C++: Tested and working
// Windows XP, MinGW g++: Tested and working
// OS X, g++: Compiles, otherwise untested
// QNX: Serial initialization routine included, otherwise untested
//
//----------------------------------------------------------------------------------


//----------------------------------------------------------------------------------
//
// Options String Format
//
// SerialCharDevice: "device,speed", e.g. "/dev/ttyS0,115200"
// TCPCharDevice: "host,port", e.g. "192.168.0.5,12345"
// TCPServerCharDevice: "host,port" or "host/netmask,port", e.g. "192.168.0.5,12345"
//   or "192.168.0.5/255.255.255.0,12345"
// UDPCharDevice: "host,port", e.g. "192.168.0.5,12345"
// UDPServerCharDevice: "host,port" or "host/netmask,port", e.g. "192.168.0.5,12345"
//   or "192.168.0.5/255.255.255.0,12345"
// CharDeviceWrapper: any of the above, with "serial", "tcp", "tcpserver", "udp", or
//   "udpserver" in front, e.g. "serial,/dev/ttyS0,115200" or "udp,192.168.0.5,12345"
//
//----------------------------------------------------------------------------------


#ifndef __CHARDEVICE_H__
#define __CHARDEVICE_H__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#endif


#ifndef WIN32
#define INVALID_HANDLE_VALUE -1
#define INVALID_SOCKET -1
typedef int HANDLE;
typedef int SOCKET;
#endif

#if defined(WIN32) && !defined(__GNUC__)
typedef int ssize_t;
#endif


typedef unsigned char  BYTE;
typedef unsigned short int WORD;
typedef unsigned int DOUBLEWORD;


// base character device class
class CharDevice
{
public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    virtual int open( bool wait = true ) = 0;
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    virtual ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true ) = 0;
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    virtual ssize_t write( const void *buf, size_t count, bool wait = true ) = 0;
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    virtual int close( void ) = 0;
    
    // get elapsed time, first call returns 0.0
    double get_time( void );
    
    
    // constructor
    CharDevice( )
    {
    }
    
    
    // destructor
    virtual ~CharDevice( )
    {
    }
};


// character device for serial communication
class SerialCharDevice : public CharDevice
{
private:

    // private copy of options string
    char *options;
    // device handle
    HANDLE fd;

public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    
    
    // constructor (_options = options string, wait = whether to wait for connection)
    SerialCharDevice( const char *_options, bool wait = true )
    {
        int retval;
    
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        fd = INVALID_HANDLE_VALUE;
        retval = this->open( wait );
        assert( retval != -1 );
    }
    
    
    // destructor
    ~SerialCharDevice( )
    {
        this->close( );
        delete[] options;
    }
};


// character device for TCP client communication
class TcpCharDevice : public CharDevice
{
private:

    // private copy of options string
    char *options;
    // device handle
    SOCKET fd;
    // device handle for connecting
    SOCKET connectsock;
    // whether currently trying to connect
    bool connecting;
    // time of last successful read
    double lastReadTime;
    #ifdef WIN32
    // whether started WSA
    bool startedWSA;
    #endif

public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    
protected:
    // close connection (returns 0 on success or -1 on error)
    int close_connection( void );
    
public:


    // constructor (_options = options string, wait = whether to wait for connection)
    TcpCharDevice( const char *_options, bool wait = true )
    {
        int retval;
    
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        connecting = false;
        fd = INVALID_SOCKET;
        connectsock = INVALID_SOCKET;
        lastReadTime = this->get_time( );
        #ifdef WIN32
        startedWSA = false;
        #endif
        retval = this->open( wait );
        assert( !wait || retval != -1 );
    }
    
    
    // destructor
    ~TcpCharDevice( )
    {
        this->close( );
        delete[] options;
    }
};


// character device for TCP server communication
class TcpServerCharDevice : public CharDevice
{
private:

    // private copy of options string
    char *options;
    // device handle
    SOCKET fd;
    // device handle for listening for connections
    SOCKET listensock;
    // whether currently listening
    bool listening;
    // time of last successful read
    double lastReadTime;
    #ifdef WIN32
    // whether started WSA
    bool startedWSA;
    #endif

public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    // close current client (returns 0 on success or -1 on error)
    int close_client( void );
    // whether currently connected
    bool connected( void );
    
    
    // constructor (_options = options string, wait = whether to wait for connection)
    TcpServerCharDevice( const char *_options, bool wait = true )
    {
        int retval;
    
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        listening = false;
        fd = INVALID_SOCKET;
        listensock = INVALID_SOCKET;
        lastReadTime = this->get_time( );
        #ifdef WIN32
        startedWSA = false;
        #endif
        retval = this->open( wait );
        assert( !wait || retval != -1 );
    }
    
    
    // destructor
    ~TcpServerCharDevice( )
    {
        this->close( );
        delete[] options;
    }
};


// character device for UDP client communication
class UdpCharDevice : public CharDevice
{
private:

    // private copy of options string
    char *options;
    // device handle
    SOCKET fd;
    // time of last successful read
    double lastReadTime;
    #ifdef WIN32
    // whether started WSA
    bool startedWSA;
    #endif

public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    
protected:
    // close connection (returns 0 on success or -1 on error)
    int close_connection( void );
    
public:

    
    // constructor (_options = options string, wait = whether to wait for connection)
    UdpCharDevice( const char *_options, bool wait = true )
    {
        int retval;
    
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        fd = INVALID_SOCKET;
        lastReadTime = this->get_time( );
        #ifdef WIN32
        startedWSA = false;
        #endif
        retval = this->open( wait );
        assert( !wait || retval != -1 );
    }
    
    
    // destructor
    ~UdpCharDevice( )
    {
        this->close( );
        delete[] options;
    }
};


// character device for UDP server communication
class UdpServerCharDevice : public CharDevice
{
private:

    // private copy of options string
    char *options;
    // device handle
    SOCKET fd;
    // address of connected client
    sockaddr_in connectedClientAddr;
    // whether address of connected client is valid
    bool connectedClientAddrValid;
    // time of last successful read
    double lastReadTime;
    #ifdef WIN32
    // whether started WSA
    bool startedWSA;
    #endif

public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    // close current client (returns 0 on success or -1 on error)
    int close_client( void );
    // whether currently connected
    bool connected( void );
    
    
    // constructor (_options = options string, wait = whether to wait for connection)
    UdpServerCharDevice( const char *_options, bool wait = true )
    {
        int retval;
    
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        connectedClientAddrValid = false;
        fd = INVALID_SOCKET;
        lastReadTime = this->get_time( );
        #ifdef WIN32
        startedWSA = false;
        #endif
        retval = this->open( wait );
        assert( !wait || retval != -1 );
    }
    
    
    // destructor
    ~UdpServerCharDevice( )
    {
        this->close( );
        delete[] options;
    }
};


// character device for communication using any of the above devices
class CharDeviceWrapper : public CharDevice
{
protected:
    
    // char device in use
    CharDevice *charDevice;
    // private copy of options string
    char *options;
    

protected:
    
    // cleanup
    void cleanup( void );
    
    
public:

    // open device, called first time by constructor (wait = whether to wait for connection,
    //   returns -1 on error)
    int open( bool wait = true );
    // read from device (buf = buffer into which to read, count = max number of bytes to read,
    //   maxInterval = time in seconds since last successful read after which to determine that
    //   link has gone down, wait = whether to wait for reconnection if link has gone down,
    //   returns number of bytes read or -1 on error)
    ssize_t read( void *buf, size_t count, double maxInterval = -1.0, bool wait = true );
    // write to device (buf = buffer from which to write, count = number number of bytes to write,
    //   wait = whether to wait for reconnection if link has gone down, returns number of bytes
    //   written or -1 on error)
    ssize_t write( const void *buf, size_t count, bool wait = true );
    // close device, called last time by destructor (returns 0 on success or -1 on error)
    int close( void );
    

    // constructor (_options = options string, wait = whether to wait for connection)
    CharDeviceWrapper( const char *_options, bool wait = true )
    {
        options = new char[ strlen( _options ) + 1 ];
        strcpy( options, _options );
        this->open( wait );
    }
    
    
    // destructor
    virtual ~CharDeviceWrapper( )
    {
    }
};


class BufferedCharDevice
{
protected:
    
    // char device in use
    CharDeviceWrapper *charDevice;
    
    // circular buffer
    int circbufSize;
    BYTE *circbuf;
    BYTE circbufStart;
    BYTE circbufNext;
    int circbufLength;
    
    
protected:

    // initialization (device = device options string, wait = whether to wait for connection)
    void init( const char *device, bool wait = true );
    // cleanup
    void cleanup( void );
    
    // read bytes into circular buffer (maxInterval = time in seconds since last successful
    //    read after which to determine that link has gone down, wait = whether to wait for
    //    reconnection if link has gone down)
    void read_into_buffer( double maxInterval = -1.0, bool wait = true );
    
    
public:

    // process input, to be implemented by descendant
    virtual void process_input( void ) = 0;
    

    // constructor
    BufferedCharDevice( )
    {
    }
    
    
    // destructor
    virtual ~BufferedCharDevice( )
    {
    }
};

#endif // __CHARDEVICE_H__
