/*****************************************************************************
Title: CRRCsim, the Charles River Radio Control Club Flight Simulator Project
Authors:
Jan Kansky:  Programming
Mark Drela:  Aerodynamics
Jan Reucker
Jens Wilhelm Wulf
 
Purpose:  The idea for CRRCsim is to take away any last excuse you might
have for not using a flight simulator to keep your thumbs certified for RC
flying.   It's free, it works, enjoy.  This is an open source project, so if
you don't like the way something works, help us fix it!
 
Thanks:
  Bruce Jackson for the LaRCsim framework.
  Flight Gear project for the sky sphere.
 
Contacts:
If you'd like to help with CRRCSIM, then send me an email!
    email                : kansky@ll.mit.edu

*****************************************************************************/

/** \file crrc_loadair.cpp
 * 
 *  Routines dealing with .air file loading
 */

#include "crrc_loadair.h"

#include <fstream>
#include <iostream>
#include <math.h>

#include <plib/ssg.h>

#include "global.h"
#include "crrc_sound.h"
#include "ls_constants.h"
#include "crrc_main.h"
#include "crrc_graphics.h"
#include "xmlmodelfile.h"
#include "crrc_animation.h"


/*****************************************************************************/

std::string getParamAsString(const char *datafile, const char *param)
{
  std::ifstream airdata(datafile, std::ios::in);
  std::string par = "";
  std::string value = "";

  if (!airdata.is_open())
  {
    std::cerr << "Unable to open airplane specification file ";
    std::cerr << datafile << std::endl;
  }
  else
  {
    bool found = false;
    while ((airdata >> par) && !found)
    {
      if (par == param)
      {
        airdata >> value;
        found = true;
      }
      // skip to end of line
      airdata.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
    }
    airdata.close();
  }
  return value;
}

bool getParamAsLong(const char *datafile, const char *param, long *value)
{
  bool ret = false;
  std::string stringval = getParamAsString(datafile, param);
  
  if (stringval != "")
  {
    *value = strtol(stringval.c_str(), NULL, 10);
    ret = true;
  }
  
  return ret;
}

/*****************************************************************************/

CRRCAirplane::CRRCAirplane()
{
  
}

CRRCAirplane::~CRRCAirplane()
{
  for (int i = 0; i < (int)sound.size(); i++)
  {
    Global::soundserver->stopChannel(sound[i]->getChannel());
    delete sound[i];
  }
}



/*****************************************************************************/

CRRCAirplaneLaRCSim::CRRCAirplaneLaRCSim()
{
}


CRRCAirplaneLaRCSim::CRRCAirplaneLaRCSim(SimpleXMLTransfer* xml)
{
  quadric = gluNewQuadric();
  
  /*
  dealloc_plane_model();
  compile_airplane();
  compile_airplane_shadow();
   */
    
  initSound(xml);    
}

// todo: remove this method when transition to xml is done
//       really?
CRRCAirplaneLaRCSim::CRRCAirplaneLaRCSim(const char* filename)
{
  FILE *airplane_data;
  char parameter_name[256]; //  Name of parameter
  char trailing_comment[256];
  double parameter_value;

  bool done_parsing_extrusion; // Hit the end_extrusion keyword
  bool done_parsing_cylinder;  // Hit the end_cylinder keyword
  bool done_parsing_sphere;    // Hit the end_sphere keyword
  bool done_parsing_triangles; // Hit the end_triangle keyword
  bool done_parsing_gear;      // Hit the end_gear keyword
  extrusion *extrusion_ptr;   // Pointer to the current extrusion being read
  extrusion *last_extrusion=NULL; //Pointer to the last filled extrusion struct
  cylinder *cylinder_ptr;         // Pointer to the current cylinder being read
  cylinder *last_cylinder=NULL;   // Pointer to the last filled cylinder struct
  sphere *sphere_ptr;             // Pointer to the current sphere being read
  sphere *last_sphere=NULL;     // Pointer to the last filled sphere struct
  triangle *triangle_ptr=NULL;  //Pointer to the current triangles struct
  triangle *last_triangle=NULL; // Pointer to the last filled triangle struct
  int loop;                     // Loop counter for extrusion parsing
  float p1,p2,p3,p4,p5,p6;      // Temporary variables used to read in file
  float x_component,y_component,magnitude;
  float dx1,dy1,dz1,dx2,dy2,dz2,normx,normy,normz,amplitude;
  double dPitchFactor = 0.002;

  // init members
  extrusions=NULL;
  cylinders=NULL;
  spheres=NULL;
  triangles=NULL;
  
  quadric = gluNewQuadric();

  // read file
  airplane_data = fopen(filename, "r");
  if (airplane_data == NULL)
  {
    std::string msg = "Unable to open airplane specification file ";
    msg += filename;

    throw std::runtime_error(msg);
  }
  else
  {
    dealloc_plane_model();

    while (fscanf(airplane_data,"%s %lf %[^\n]",(char *)parameter_name,
                  &parameter_value,trailing_comment)!=EOF)
    {
#if (DEBUG_READ_AIRPLANE_DATA != 0)
      printf("%s %lf %s\n",parameter_name, parameter_value, trailing_comment);
#endif

      if (strcasecmp(parameter_name,"max_thrust")==0)
        max_thrust=parameter_value;
      else if (strcasecmp(parameter_name,"engine_sound_pitchfactor")==0)
        dPitchFactor=parameter_value;
      else if (strcasecmp(parameter_name,"gear")==0)
      {
        printf("In gear\n");
        done_parsing_gear=false;
        while(!done_parsing_gear)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          if (strcasecmp(parameter_name,"end_gear")==0)
          {
            done_parsing_gear=true;
          }
        }

      }
      else if (strcasecmp(parameter_name,"triangles")==0)
      {
        printf("In triangles\n");
        triangle_ptr=(triangle*)malloc(sizeof(triangle));
        triangle_ptr->triangle_name=(char*)malloc((strlen(trailing_comment)+1)*
                                                  sizeof(char));
        strcpy(triangle_ptr->triangle_name,trailing_comment);
        triangle_ptr->num_triangles=(unsigned long int)parameter_value;
        triangle_ptr->vertices=(GLfloat **)malloc(triangle_ptr->
                                                  num_triangles*sizeof(GLfloat *));
        triangle_ptr->normals=(GLfloat **)malloc(triangle_ptr->
                                                 num_triangles*sizeof(GLfloat *));
        triangle_ptr->vertices[0]=(GLfloat *)malloc(triangle_ptr->
                                                    num_triangles*sizeof(GLfloat)*9);
        triangle_ptr->normals[0]=(GLfloat *)malloc(triangle_ptr->
                                                   num_triangles*sizeof(GLfloat)*3);
        for (loop=1;loop<(int)triangle_ptr->num_triangles;loop++)
        {
          triangle_ptr->vertices[loop]=triangle_ptr->vertices[0]+ loop*9;
          triangle_ptr->normals[loop]=triangle_ptr->normals[0]+ loop*3;
        }
        triangle_ptr->next_triangle=NULL;
        if (last_triangle==NULL)
          triangles=triangle_ptr;
        else
          last_triangle->next_triangle=triangle_ptr;
        last_triangle=triangle_ptr;
        done_parsing_triangles=false;
        while(!done_parsing_triangles)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);
#if (DEBUG_READ_AIRPLANE_DATA != 0)

          printf("     %s\n",parameter_name);
#endif

          if (strcasecmp(parameter_name,"translate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            triangle_ptr->translate_x=p1;
            triangle_ptr->translate_y=p2;
            triangle_ptr->translate_z=p3;
          }
          else if (strcasecmp(parameter_name,"rotate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            triangle_ptr->rotate_x=p1;
            triangle_ptr->rotate_y=p2;
            triangle_ptr->rotate_z=p3;
          }
          else if (strcasecmp(parameter_name,"color")==0)
          {
            fscanf(airplane_data,"%f %f %f %f",&p1,&p2,&p3,&p4);
            triangle_ptr->colors[0]=p1;
            triangle_ptr->colors[1]=p2;
            triangle_ptr->colors[2]=p3;
            triangle_ptr->colors[3]=p4;
          }
          else if (strcasecmp(parameter_name,"vertices")==0)
          {
            for (loop=0;loop<(int)triangle_ptr->num_triangles;loop++)
            {
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              triangle_ptr->vertices[loop][0]=p1;
              triangle_ptr->vertices[loop][1]=p2;
              triangle_ptr->vertices[loop][2]=p3;
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              triangle_ptr->vertices[loop][3]=p1;
              triangle_ptr->vertices[loop][4]=p2;
              triangle_ptr->vertices[loop][5]=p3;
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              triangle_ptr->vertices[loop][6]=p1;
              triangle_ptr->vertices[loop][7]=p2;
              triangle_ptr->vertices[loop][8]=p3;

              dx1=triangle_ptr->vertices[loop][3]-
                triangle_ptr->vertices[loop][0];
              dy1=triangle_ptr->vertices[loop][1]-
                triangle_ptr->vertices[loop][4];
              dz1=triangle_ptr->vertices[loop][2]-
                triangle_ptr->vertices[loop][5];
              dx2=triangle_ptr->vertices[loop][3]-
                triangle_ptr->vertices[loop][6];
              dy2=triangle_ptr->vertices[loop][4]-
                triangle_ptr->vertices[loop][7];
              dz2=triangle_ptr->vertices[loop][5]-
                triangle_ptr->vertices[loop][8];
              normx=dy1*dz2-dz1*dy2;
              normy=dz1*dx2-dx1*dz2;
              normz=dx1*dy2-dy1*dx2;
              amplitude=sqrt(normx*normx+normy*normy+normz*normz);
              triangle_ptr->normals[loop][0]=normx/amplitude;
              triangle_ptr->normals[loop][1]=normy/amplitude;
              triangle_ptr->normals[loop][2]=normz/amplitude;
            }
          }
          else if (strcasecmp(parameter_name,"end_triangles")==0)
          {
            done_parsing_triangles=true;
          }
        }
      }
      else if (strcasecmp(parameter_name,"sphere")==0)
      {
        printf("In sphere\n");
        sphere_ptr=(sphere*)malloc(sizeof(sphere));
        sphere_ptr->sphere_name=(char*)malloc((strlen(trailing_comment)+1)*
                                              sizeof(char));
        strcpy(sphere_ptr->sphere_name,trailing_comment);
        sphere_ptr->next_sphere=NULL;
        if (last_sphere==NULL)
          spheres=sphere_ptr;
        else
          last_sphere->next_sphere=sphere_ptr;
        last_sphere=sphere_ptr;
        done_parsing_sphere=false;
        while(!done_parsing_sphere)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);
#if (DEBUG_READ_AIRPLANE_DATA != 0)

          printf("     %s\n",parameter_name);
#endif

          if (strcasecmp(parameter_name,"translate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            sphere_ptr->translate_x=p1;
            sphere_ptr->translate_y=p2;
            sphere_ptr->translate_z=p3;
          }
          else if (strcasecmp(parameter_name,"rotate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            sphere_ptr->rotate_x=p1;
            sphere_ptr->rotate_y=p2;
            sphere_ptr->rotate_z=p3;
          }
          else if (strcasecmp(parameter_name,"dimension")==0)
          {
            fscanf(airplane_data,"%f",&p1);
            sphere_ptr->radius=p1;
          }
          else if (strcasecmp(parameter_name,"grid")==0)
          {
            fscanf(airplane_data,"%f %f",&p1,&p2);
            sphere_ptr->slices=(int)p1;
            sphere_ptr->stacks=(int)p2;
          }
          else if (strcasecmp(parameter_name,"color")==0)
          {
            fscanf(airplane_data,"%f %f %f %f",&p1,&p2,&p3,&p4);
            sphere_ptr->colors[0]=p1;
            sphere_ptr->colors[1]=p2;
            sphere_ptr->colors[2]=p3;
            sphere_ptr->colors[3]=p4;
          }
          else if (strcasecmp(parameter_name,"end_sphere")==0)
          {
            done_parsing_sphere=true;
          }
        }
      }
      else if (strcasecmp(parameter_name,"cylinder")==0)
      {
        printf("In cylinder\n");
        cylinder_ptr=(cylinder*)malloc(sizeof(cylinder));
        cylinder_ptr->cylinder_name=(char*)malloc((strlen(trailing_comment)+1)*
                                                  sizeof(char));
        strcpy(cylinder_ptr->cylinder_name,trailing_comment);
        cylinder_ptr->next_cylinder=NULL;
        if (last_cylinder==NULL)
          cylinders=cylinder_ptr;
        else
          last_cylinder->next_cylinder=cylinder_ptr;
        last_cylinder=cylinder_ptr;
        done_parsing_cylinder=false;
        while(!done_parsing_cylinder)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);
#if (DEBUG_READ_AIRPLANE_DATA != 0)

          printf("     %s\n",parameter_name);
#endif

          if (strcasecmp(parameter_name,"translate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            cylinder_ptr->translate_x=p1;
            cylinder_ptr->translate_y=p2;
            cylinder_ptr->translate_z=p3;
          }
          else if (strcasecmp(parameter_name,"rotate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            cylinder_ptr->rotate_x=p1;
            cylinder_ptr->rotate_y=p2;
            cylinder_ptr->rotate_z=p3;
          }
          else if (strcasecmp(parameter_name,"dimension")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            cylinder_ptr->radius_1=p1;
            cylinder_ptr->radius_2=p2;
            cylinder_ptr->length=p3;
          }
          else if (strcasecmp(parameter_name,"grid")==0)
          {
            fscanf(airplane_data,"%f %f",&p1,&p2);
            cylinder_ptr->radial_sampling=(int)p1;
            cylinder_ptr->longitudinal_sampling=(int)p2;
          }
          else if (strcasecmp(parameter_name,"color")==0)
          {
            fscanf(airplane_data,"%f %f %f %f",&p1,&p2,&p3,&p4);
            cylinder_ptr->colors[0]=p1;
            cylinder_ptr->colors[1]=p2;
            cylinder_ptr->colors[2]=p3;
            cylinder_ptr->colors[3]=p4;
          }
          else if (strcasecmp(parameter_name,"end_cylinder")==0)
          {
            done_parsing_cylinder=true;
          }
        }
      }
      else if (strcasecmp(parameter_name,"extrusion")==0)
      {
        printf("In extrusion.\n");
        extrusion_ptr=(extrusion*)malloc(sizeof(extrusion));
        extrusion_ptr->num_points_on_path=(int)parameter_value+2;
        extrusion_ptr->extrusion_name=(char*)malloc((strlen(trailing_comment)+1)*
                                                    sizeof(char));
        strcpy(extrusion_ptr->extrusion_name,trailing_comment);
        extrusion_ptr->point_array=(gleDouble **)malloc(extrusion_ptr->
                                                        num_points_on_path*sizeof(gleDouble *));
        extrusion_ptr->point_array[0]=(gleDouble *)malloc(3*extrusion_ptr->
                                                          num_points_on_path*sizeof(gleDouble));
        extrusion_ptr->colors=(GLfloat **)malloc(extrusion_ptr->
                                                 num_points_on_path*sizeof(GLfloat *));
        extrusion_ptr->colors[0]=(GLfloat *)malloc(3*extrusion_ptr->
                                                   num_points_on_path*sizeof(GLfloat));
        extrusion_ptr->scaling=(gleAffine *)malloc(extrusion_ptr->
                                                   num_points_on_path*sizeof(gleAffine));
        for (loop=1;loop<extrusion_ptr->num_points_on_path;loop++)
        {
          extrusion_ptr->point_array[loop]=extrusion_ptr->point_array[0]+
            loop*3;
          extrusion_ptr->colors[loop]=extrusion_ptr->colors[0]+
            loop*3;
        }
        // Defaults for rotate:
        extrusion_ptr->rotate_x = 0;
        extrusion_ptr->rotate_y = 0;
        extrusion_ptr->rotate_z = 0;
        // assume that there is no next extrusion:
        extrusion_ptr->next_extrusion=NULL;
        if (last_extrusion==NULL)    // If this is the first parsed extrusion
          extrusions=extrusion_ptr;
        else
          last_extrusion->next_extrusion=extrusion_ptr;
        last_extrusion=extrusion_ptr;
        done_parsing_extrusion=false;
        while(!done_parsing_extrusion)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);
#if (DEBUG_READ_AIRPLANE_DATA != 0)

          printf("     %s\n",parameter_name);
#endif

          if (strcasecmp(parameter_name,"translate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            extrusion_ptr->translate_x=p1;
            extrusion_ptr->translate_y=p2;
            extrusion_ptr->translate_z=p3;
          }
          else if (strcasecmp(parameter_name,"rotate")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            extrusion_ptr->rotate_x=p1;
            extrusion_ptr->rotate_y=p2;
            extrusion_ptr->rotate_z=p3;
          }
          else if (strcasecmp(parameter_name,"up_vector")==0)
          {
            fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
            extrusion_ptr->up[0]=p1;
            extrusion_ptr->up[1]=p2;
            extrusion_ptr->up[2]=p3;
          }
          else if (strcasecmp(parameter_name,"path")==0)
          {
            for (loop=1;loop<extrusion_ptr->num_points_on_path-1;loop++)
            {
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              extrusion_ptr->point_array[loop][0]=p1;
              extrusion_ptr->point_array[loop][1]=p2;
              extrusion_ptr->point_array[loop][2]=p3;
            }
            extrusion_ptr->point_array[0][0]=extrusion_ptr->
              point_array[1][0];
            extrusion_ptr->point_array[0][1]=extrusion_ptr->
              point_array[1][1];
            extrusion_ptr->point_array[0][2]=extrusion_ptr->
              point_array[1][2];
            extrusion_ptr->point_array[extrusion_ptr->
                                       num_points_on_path-1][0]=extrusion_ptr->
              point_array[extrusion_ptr->num_points_on_path-2][0];
            extrusion_ptr->point_array[extrusion_ptr->
                                       num_points_on_path-1][1]=extrusion_ptr->
              point_array[extrusion_ptr->num_points_on_path-2][1];
            extrusion_ptr->point_array[extrusion_ptr->
                                       num_points_on_path-1][2]=extrusion_ptr->
              point_array[extrusion_ptr->num_points_on_path-2][2];
          }
          else if (strcasecmp(parameter_name,"colors")==0)
          {
            for (loop=1;loop<extrusion_ptr->num_points_on_path-1;loop++)
            {
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              extrusion_ptr->colors[loop][0]=p1;
              extrusion_ptr->colors[loop][1]=p2;
              extrusion_ptr->colors[loop][2]=p3;
            }
            extrusion_ptr->colors[0][0]=extrusion_ptr->
              colors[1][0];
            extrusion_ptr->colors[0][1]=extrusion_ptr->
              colors[1][1];
            extrusion_ptr->colors[0][2]=extrusion_ptr->
              colors[1][2];
            extrusion_ptr->colors[extrusion_ptr->
                                  num_points_on_path-1][0]=extrusion_ptr->
              colors[extrusion_ptr->num_points_on_path-2][0];
            extrusion_ptr->colors[extrusion_ptr->
                                  num_points_on_path-1][1]=extrusion_ptr->
              colors[extrusion_ptr->num_points_on_path-2][1];
            extrusion_ptr->colors[extrusion_ptr->
                                  num_points_on_path-1][2]=extrusion_ptr->
              colors[extrusion_ptr->num_points_on_path-2][2];
          }
          else if (strcasecmp(parameter_name,"scaling")==0)
          {
            for (loop=1;loop<extrusion_ptr->num_points_on_path-1;loop++)
            {
              fscanf(airplane_data,"%f %f %f %f %f %f",&p1,&p2,&p3,&p4,
                     &p5,&p6);
              extrusion_ptr->scaling[loop][0][0]=p1;
              extrusion_ptr->scaling[loop][0][1]=p2;
              extrusion_ptr->scaling[loop][0][2]=p3;
              extrusion_ptr->scaling[loop][1][0]=p4;
              extrusion_ptr->scaling[loop][1][1]=p5;
              extrusion_ptr->scaling[loop][1][2]=p6;
            }
            extrusion_ptr->scaling[0][0][0]=extrusion_ptr->
              scaling[1][0][0];
            extrusion_ptr->scaling[0][0][1]=extrusion_ptr->
              scaling[1][0][1];
            extrusion_ptr->scaling[0][0][2]=extrusion_ptr->
              scaling[1][0][2];
            extrusion_ptr->scaling[0][1][0]=extrusion_ptr->
              scaling[1][1][0];
            extrusion_ptr->scaling[0][1][1]=extrusion_ptr->
              scaling[1][1][1];
            extrusion_ptr->scaling[0][1][2]=extrusion_ptr->
              scaling[1][1][2];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][0][0]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][0][0];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][0][1]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][0][1];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][0][2]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][0][2];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][1][0]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][1][0];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][1][1]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][1][1];
            extrusion_ptr->scaling[extrusion_ptr->
                                   num_points_on_path-1][1][2]=extrusion_ptr->
              scaling[extrusion_ptr->num_points_on_path-2][1][2];
          }
          else if (strcasecmp(parameter_name,"section")==0)
          {
            fscanf(airplane_data,"%f",&p1);
            extrusion_ptr->num_contour_elements=(int)p1;
            extrusion_ptr->contour=(gleDouble **)malloc(extrusion_ptr->
                                                        num_contour_elements*sizeof(gleDouble *));
            extrusion_ptr->contour[0]=(gleDouble *)malloc(extrusion_ptr->
                                                          num_contour_elements*2*sizeof(gleDouble));
            extrusion_ptr->contour_normal_vectors=(gleDouble **)
              malloc(extrusion_ptr->num_contour_elements*
                     sizeof(gleDouble *));
            extrusion_ptr->contour_normal_vectors[0]=(gleDouble *)
              malloc(extrusion_ptr->num_contour_elements*2*
                     sizeof(gleDouble));

            for (loop=1;loop<extrusion_ptr->num_contour_elements;loop++)
            {
              extrusion_ptr->contour[loop]=extrusion_ptr->contour[0]+
                2*loop;
              extrusion_ptr->contour_normal_vectors[loop]=
                extrusion_ptr->contour_normal_vectors[0]+
                2*loop;
            }
            for (loop=0;loop<extrusion_ptr->num_contour_elements;loop++)
            {
              fscanf(airplane_data,"%f %f",&p1,&p2);
              extrusion_ptr->contour[loop][0]=p1;
              extrusion_ptr->contour[loop][1]=p2;
            }
            for (loop=2;loop<extrusion_ptr->num_contour_elements;loop++)
            {
              x_component=(extrusion_ptr->contour[loop][0]-
                           extrusion_ptr->contour[loop-1][0]);
              y_component=(extrusion_ptr->contour[loop][1]-
                           extrusion_ptr->contour[loop-1][1]);
              magnitude=sqrt(x_component*x_component+
                             y_component*y_component);
              extrusion_ptr->contour_normal_vectors[loop-1][0]=
                y_component/magnitude;
              extrusion_ptr->contour_normal_vectors[loop-1][1]=
                -1*x_component/magnitude;

              x_component=(extrusion_ptr->contour[loop-1][0]-
                           extrusion_ptr->contour[loop-2][0]);
              y_component=(extrusion_ptr->contour[loop-1][1]-
                           extrusion_ptr->contour[loop-2][1]);
              magnitude=sqrt(x_component*x_component+
                             y_component*y_component);

              extrusion_ptr->contour_normal_vectors[loop-1][0]+=
                y_component/magnitude;
              extrusion_ptr->contour_normal_vectors[loop-1][1]+=
                -1*x_component/magnitude;
              magnitude=sqrt(
                             extrusion_ptr->contour_normal_vectors[loop-1][0]*
                             extrusion_ptr->contour_normal_vectors[loop-1][0]+
                             extrusion_ptr->contour_normal_vectors[loop-1][1]*
                             extrusion_ptr->contour_normal_vectors[loop-1][1]);
              extrusion_ptr->contour_normal_vectors[loop-1][0]/=
                magnitude;
              extrusion_ptr->contour_normal_vectors[loop-1][1]/=
                magnitude;
            }
            x_component=(extrusion_ptr->contour[1][0]-
                         extrusion_ptr->contour[0][0]);
            y_component=(extrusion_ptr->contour[1][1]-
                         extrusion_ptr->contour[0][1]);
            magnitude=sqrt(x_component*x_component+
                           y_component*y_component);
            extrusion_ptr->contour_normal_vectors[0][0]=y_component/
              magnitude;
            extrusion_ptr->contour_normal_vectors[0][1]= -1*x_component/
              magnitude;

            x_component=(extrusion_ptr->contour[extrusion_ptr->
                                                num_contour_elements-1][0]-
                         extrusion_ptr->contour[extrusion_ptr->
                                                num_contour_elements-2][0]);
            y_component=(extrusion_ptr->contour[extrusion_ptr->
                                                num_contour_elements-1][1]-
                         extrusion_ptr->contour[extrusion_ptr->
                                                num_contour_elements-2][1]);
            magnitude=sqrt(x_component*x_component+
                           y_component*y_component);
            extrusion_ptr->contour_normal_vectors[extrusion_ptr->
                                                  num_contour_elements-1][0]=y_component/
              magnitude;
            extrusion_ptr->contour_normal_vectors[extrusion_ptr->
                                                  num_contour_elements-1][1]= -1*x_component/
              magnitude;
          }
          else if (strcasecmp(parameter_name,"end_extrusion")==0)
          {
            done_parsing_extrusion=true;
          }
        }
      }
    }
    compile_airplane();
    compile_airplane_shadow();
    initSound(filename, dPitchFactor);
    
    fclose(airplane_data);
  }
}

CRRCAirplaneLaRCSim::~CRRCAirplaneLaRCSim()
{
  gluDeleteQuadric(quadric);
}


void CRRCAirplaneLaRCSim::draw(CRRC_AirplaneSim* airplane)
{
  float X_cg_rwy =    airplane->getPos().r[0];
  float Y_cg_rwy =    airplane->getPos().r[1];
  float H_cg_rwy = -1*airplane->getPos().r[2];
  
  float Psi   =    airplane->getPsi();
  float Theta =    airplane->getTheta();
  float Phi   =    airplane->getPhi();
  
  glPushMatrix();
#if COOR_NEW == 1
  glTranslatef(Y_cg_rwy,H_cg_rwy, -1*X_cg_rwy);
  glRotatef(180-Psi*RAD_TO_DEG,0,1,0);
  glRotatef(   Theta*RAD_TO_DEG,-1,0,0); 
#else
  glTranslatef(X_cg_rwy,H_cg_rwy, Y_cg_rwy);
  glRotatef(90,0,1,0);
  glRotatef(Psi*RAD_TO_DEG,0,-1,0);
  glRotatef(-1*Theta*RAD_TO_DEG,-1,0,0);
#endif
  glRotatef(Phi*RAD_TO_DEG,0,0,1);

  if (Global::training_mode)
  {
    if (airplane->isStalling())
    {
      glPushMatrix();
      glColor4f(1,0,0,1);
      gluSphere(quadric,0.42,10,10);
      glPopMatrix();
    }
  }

  glCallList(airplane_list);

  glPopMatrix();
}

void CRRCAirplaneLaRCSim::draw_shadow(CRRC_AirplaneSim* airplane,
                                      float             shadow_matrix[4][4])
{
  float X_cg_rwy =    airplane->getPos().r[0];
  float Y_cg_rwy =    airplane->getPos().r[1];
  float H_cg_rwy = -1*airplane->getPos().r[2];
  
  float Psi   =    airplane->getPsi();
  float Theta =    airplane->getTheta();
  float Phi   =    airplane->getPhi();
  
  GLfloat shadow_ambient[] = {0.0, 0.0, 0.0, 1.0};
  GLfloat shadow_diffuse[] = {0.0, 0.0, 0.0, 0.3};
  GLfloat shadow_shiny[] = {0.0};
  GLfloat shadow_specular[] = {0.0, 0.0, 0.0, 1.0};

  glMaterialfv(GL_FRONT, GL_AMBIENT, shadow_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, shadow_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, shadow_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, shadow_shiny);
  glColor4f(0,0,0,1.0);
  glPushMatrix();
  glMultMatrixf((const GLfloat *) shadow_matrix);
  
#if COOR_NEW == 1
  glTranslatef(Y_cg_rwy,H_cg_rwy, -1*X_cg_rwy);
  glRotatef(180-Psi*RAD_TO_DEG,0,1,0);
  glRotatef(   Theta*RAD_TO_DEG,-1,0,0); 
#else
  glTranslatef(X_cg_rwy,H_cg_rwy, Y_cg_rwy);
  glRotatef(90,0,1,0);
  glRotatef(Psi*RAD_TO_DEG,0,-1,0);
  glRotatef(-1*Theta*RAD_TO_DEG,-1,0,0);
#endif      
  glRotatef(Phi*RAD_TO_DEG,0,0,1);

  glCallList(airplane_shadow_list);

  glPopMatrix();
}

void CRRCAirplaneLaRCSim::dealloc_plane_model()
{
  // de-allocate list of extrusions
  extrusion *extrusion_ptr = extrusions;
  extrusion *last_extrusion_ptr;
  
  while (extrusion_ptr != NULL)
  {
    free(extrusion_ptr->extrusion_name);
    free(extrusion_ptr->scaling);
    
    free(extrusion_ptr->point_array[0]);
    free(extrusion_ptr->contour[0]);
    free(extrusion_ptr->colors[0]);
    
    free(extrusion_ptr->point_array);
    free(extrusion_ptr->contour);
    free(extrusion_ptr->colors);

    last_extrusion_ptr = extrusion_ptr;
    extrusion_ptr = (extrusion*)extrusion_ptr->next_extrusion;
    free(last_extrusion_ptr);
  }

  // de-allocate list of triangles
  triangle *triangle_ptr = triangles;
  triangle *last_triangle_ptr;

  while (triangle_ptr != NULL)
  {
    free(triangle_ptr->triangle_name);
    free(triangle_ptr->vertices[0]);
    free(triangle_ptr->vertices);
    free(triangle_ptr->normals[0]);
    free(triangle_ptr->normals);
    
    last_triangle_ptr = triangle_ptr;
    triangle_ptr = (triangle*)triangle_ptr->next_triangle;
    free(last_triangle_ptr);
  }
  
  // de-allocate list of cylinders
  cylinder *cylinder_ptr = cylinders;
  cylinder *last_cylinder_ptr;
  
  while (cylinder_ptr != NULL)
  {
    free(cylinder_ptr->cylinder_name);
    last_cylinder_ptr = cylinder_ptr;
    cylinder_ptr = (cylinder*)cylinder_ptr->next_cylinder;
    free(last_cylinder_ptr);
  }
  
  // de-allocate list of spheres
  sphere *sphere_ptr = spheres;
  sphere *last_sphere_ptr;
  
  while (sphere_ptr != NULL)
  {
    free(sphere_ptr->sphere_name);
    last_sphere_ptr = sphere_ptr;
    sphere_ptr = (sphere*)sphere_ptr->next_sphere;
    free(last_sphere_ptr);
  }

  // clean up
  extrusions = NULL;
  cylinders = NULL;
  triangles = NULL;
  spheres = NULL;
}

void CRRCAirplaneLaRCSim::compile_airplane()
{
  extrusion *extrusion_ptr;
  cylinder *cylinder_ptr;
  sphere *sphere_ptr;
  triangle *triangle_ptr;
  int loop;

  extrusion_ptr=extrusions;
  cylinder_ptr=cylinders;
  sphere_ptr=spheres;
  triangle_ptr=triangles;

  if (glIsList(airplane_list))
  {
    glDeleteLists(airplane_list, 1);
  }
  airplane_list = glGenLists(1);
  glNewList(airplane_list, GL_COMPILE);

  glEnable(GL_COLOR_MATERIAL);
  while (extrusion_ptr != NULL)
  {
    glPushMatrix();
    glTranslatef(extrusion_ptr->translate_x,extrusion_ptr->translate_y,
                 extrusion_ptr->translate_z);
    glRotatef(extrusion_ptr->rotate_x,1,0,0);
    glRotatef(extrusion_ptr->rotate_y,0,1,0);
    glRotatef(extrusion_ptr->rotate_z,0,0,1);
    gleSuperExtrusion(extrusion_ptr->num_contour_elements,
                      (gleDouble (*) [2])(*(extrusion_ptr->contour)),
                      (gleDouble (*) [2])(*(extrusion_ptr->
                                            contour_normal_vectors)),extrusion_ptr->up,
                      extrusion_ptr->
                      num_points_on_path,(gleDouble (*) [3])
                      (*(extrusion_ptr->point_array)),(float (*) [3])
                      (*(extrusion_ptr->colors)),(gleAffine (*))
                      (*(extrusion_ptr->scaling)));
    glPopMatrix();
    extrusion_ptr=(extrusion*)extrusion_ptr->next_extrusion;
  }
  glDisable(GL_COLOR_MATERIAL);
  while(cylinder_ptr!=NULL)
  {
    glPushMatrix();
    glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,
                 cylinder_ptr->colors);
    glTranslatef(cylinder_ptr->translate_x,cylinder_ptr->translate_y,
                 cylinder_ptr->translate_z);
    glRotatef(cylinder_ptr->rotate_x,1,0,0);
    glRotatef(cylinder_ptr->rotate_y,0,1,0);
    glRotatef(cylinder_ptr->rotate_z,0,0,1);
    gluCylinder(quadric,cylinder_ptr->radius_1,cylinder_ptr->radius_2,
                cylinder_ptr->length,(int)cylinder_ptr->radial_sampling,
                (int)cylinder_ptr->longitudinal_sampling);
    glPopMatrix();
    cylinder_ptr=(cylinder*)cylinder_ptr->next_cylinder;
  }
  while(sphere_ptr!=NULL)
  {
    glPushMatrix();
    glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,
                 sphere_ptr->colors);
    glTranslatef(sphere_ptr->translate_x,sphere_ptr->translate_y,
                 sphere_ptr->translate_z);
    glRotatef(sphere_ptr->rotate_x,1,0,0);
    glRotatef(sphere_ptr->rotate_y,0,1,0);
    glRotatef(sphere_ptr->rotate_z,0,0,1);
    gluSphere(quadric,sphere_ptr->radius,(int)sphere_ptr->slices,
              (int)sphere_ptr->stacks);
    glPopMatrix();
    sphere_ptr=(sphere*)sphere_ptr->next_sphere;
  }

  glEnable(GL_CULL_FACE);
  while (triangle_ptr!=NULL)
  {
    glPushMatrix();
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,
                 triangle_ptr->colors);
    glTranslatef(triangle_ptr->translate_x,triangle_ptr->translate_y,
                 triangle_ptr->translate_z);
    glRotatef(triangle_ptr->rotate_x,1,0,0);
    glRotatef(triangle_ptr->rotate_y,0,1,0);
    glRotatef(triangle_ptr->rotate_z,0,0,1);
    glBegin(GL_TRIANGLES);
    for (loop=0;loop<(int)triangle_ptr->num_triangles;loop++)
    {
      glNormal3f(triangle_ptr->normals[loop][0],
                 triangle_ptr->normals[loop][1],
                 triangle_ptr->normals[loop][2]);
      glVertex3f(triangle_ptr->vertices[loop][0],
                 triangle_ptr->vertices[loop][1],
                 triangle_ptr->vertices[loop][2]);
      glVertex3f(triangle_ptr->vertices[loop][3],
                 triangle_ptr->vertices[loop][4],
                 triangle_ptr->vertices[loop][5]);
      glVertex3f(triangle_ptr->vertices[loop][6],
                 triangle_ptr->vertices[loop][7],
                 triangle_ptr->vertices[loop][8]);
    }
    glEnd();
    triangle_ptr=(triangle*)triangle_ptr->next_triangle;
    glPopMatrix();
  }
  glDisable(GL_CULL_FACE);
  glEndList();
}

void CRRCAirplaneLaRCSim::compile_airplane_shadow()
{
  extrusion *extrusion_ptr;
  cylinder *cylinder_ptr;
  sphere  *sphere_ptr;
  triangle *triangle_ptr;
  int loop;

  if (glIsList(airplane_shadow_list))
  {
    glDeleteLists(airplane_shadow_list, 1);
  }
  airplane_shadow_list = glGenLists(1);
  glNewList(airplane_shadow_list, GL_COMPILE);

  extrusion_ptr=extrusions;
  cylinder_ptr=cylinders;
  sphere_ptr=spheres;
  triangle_ptr=triangles;
  while (extrusion_ptr != NULL)
  {
    glPushMatrix();
    glTranslatef(extrusion_ptr->translate_x,extrusion_ptr->translate_y,
                 extrusion_ptr->translate_z);
    glRotatef(extrusion_ptr->rotate_x,1,0,0);
    glRotatef(extrusion_ptr->rotate_y,0,1,0);
    glRotatef(extrusion_ptr->rotate_z,0,0,1);
    gleSuperExtrusion(extrusion_ptr->num_contour_elements,
                      (gleDouble (*) [2])(*(extrusion_ptr->contour)),
                      (gleDouble (*) [2])(*(extrusion_ptr->
                                            contour_normal_vectors)),extrusion_ptr->up,
                      extrusion_ptr->
                      num_points_on_path,(gleDouble (*) [3])
                      (*(extrusion_ptr->point_array)),(float (*) [3])
                      (*(extrusion_ptr->colors)),(gleAffine (*))
                      (*(extrusion_ptr->scaling)));
    glPopMatrix();
    extrusion_ptr=(extrusion*)extrusion_ptr->next_extrusion;
  }
  while(cylinder_ptr!=NULL)
  {
    glPushMatrix();
    glTranslatef(cylinder_ptr->translate_x,cylinder_ptr->translate_y,
                 cylinder_ptr->translate_z);
    glRotatef(cylinder_ptr->rotate_x,1,0,0);
    glRotatef(cylinder_ptr->rotate_y,0,1,0);
    glRotatef(cylinder_ptr->rotate_z,0,0,1);
    gluCylinder(quadric,cylinder_ptr->radius_1,cylinder_ptr->radius_2,
                cylinder_ptr->length,(int)cylinder_ptr->radial_sampling,
                (int)cylinder_ptr->longitudinal_sampling);
    glPopMatrix();
    cylinder_ptr=(cylinder*)cylinder_ptr->next_cylinder;
  }
  while(sphere_ptr!=NULL)
  {
    glPushMatrix();
    glTranslatef(sphere_ptr->translate_x,sphere_ptr->translate_y,
                 sphere_ptr->translate_z);
    glRotatef(sphere_ptr->rotate_x,1,0,0);
    glRotatef(sphere_ptr->rotate_y,0,1,0);
    glRotatef(sphere_ptr->rotate_z,0,0,1);
    gluSphere(quadric,sphere_ptr->radius,(int)sphere_ptr->slices,
              (int)sphere_ptr->stacks);
    glPopMatrix();
    sphere_ptr=(sphere*)sphere_ptr->next_sphere;
  }

  glEnable(GL_CULL_FACE);
  while (triangle_ptr!=NULL)
  {
    glPushMatrix();
    glTranslatef(triangle_ptr->translate_x,triangle_ptr->translate_y,
                 triangle_ptr->translate_z);
    glRotatef(triangle_ptr->rotate_x,1,0,0);
    glRotatef(triangle_ptr->rotate_y,0,1,0);
    glRotatef(triangle_ptr->rotate_z,0,0,1);
    glBegin(GL_TRIANGLES);
    for (loop=0;loop<(int)triangle_ptr->num_triangles;loop++)
    {
      glNormal3f(triangle_ptr->normals[loop][0],
                 triangle_ptr->normals[loop][1],
                 triangle_ptr->normals[loop][2]);
      glVertex3f(triangle_ptr->vertices[loop][0],
                 triangle_ptr->vertices[loop][1],
                 triangle_ptr->vertices[loop][2]);
      glVertex3f(triangle_ptr->vertices[loop][3],
                 triangle_ptr->vertices[loop][4],
                 triangle_ptr->vertices[loop][5]);
      glVertex3f(triangle_ptr->vertices[loop][6],
                 triangle_ptr->vertices[loop][7],
                 triangle_ptr->vertices[loop][8]);
    }
    glEnd();
    triangle_ptr=(triangle*)triangle_ptr->next_triangle;
    glPopMatrix();
  }
  glDisable(GL_CULL_FACE);


  glEndList();
}

void CRRCAirplaneLaRCSim::initSound(SimpleXMLTransfer* xml)
{
  SimpleXMLTransfer* cfg = XMLModelFile::getConfig(xml);
  SimpleXMLTransfer* sndcfg = cfg->getChild("sound", true);
  int children = sndcfg->getChildCount();
  int units = sndcfg->getInt("units", 0);
  
  for (int i = 0; i < children; i++)
  {
    SimpleXMLTransfer *child = sndcfg->getChildAt(i);
    std::string name = child->getName();
    
    if (name.compare("sample") == 0)
    {
      T_AirplaneSound *sample;

      // assemble relative path
      std::string soundfile;
      soundfile           = child->attribute("filename");

      // other sound attributes
      int sound_type      = child->getInt("type", SOUND_TYPE_GLIDER);
      double dPitchFactor = child->getDouble("pitchfactor", 0.002);
      double dMaxVolume   = child->getDouble("maxvolume", 1.0);
  
      if (dMaxVolume < 0.0)
      {
        dMaxVolume = 0.0;
      }
      else if (dMaxVolume > 1.0)
      {
        dMaxVolume = 1.0;
      }

  //~ if (cfg->indexOfChild("power") < 0)
    //~ max_thrust = 0;
  //~ else
    //~ max_thrust = 1;
  
      if (soundfile != "")
      {
        // Get full path (considering search paths). 
        soundfile = T_Config::getDataPath("sounds/" + soundfile);
      }
      
      // File ok? Use default otherwise.
      if (!T_Config::fileExists(soundfile))
        soundfile = T_Config::getDataPath("sounds/fan.wav");
    
      std::cout << "soundfile: " << soundfile << "\n";
      //~ std::cout << "max_thrust: " << max_thrust << "\n";
      std::cout << "soundserver: " << Global::soundserver << "\n";
  
      // Only make noise if a sound file is available
      if (soundfile != "" && Global::soundserver != (CRRCAudioServer*)0)
      {        
        std::cout << "Using airplane sound " << soundfile << ", type " << sound_type << ", max vol " << dMaxVolume << std::endl;
        
        if (sound_type == SOUND_TYPE_GLIDER)
        {
          T_GliderSound *glidersound;
          float flMinRelV, flMaxRelV, flMaxDist;
          flMinRelV = (float)child->getDouble("v_min", 1.5);
          flMaxRelV = (float)child->getDouble("v_max", 4.0);
          flMaxDist = (float)child->getDouble("dist_max", 300);
          
          if (units == 1)
          {
            // convert from metric units to ft.
            flMaxDist *= METERS_TO_FT;
          }
          
          glidersound = new T_GliderSound(soundfile.c_str(), Global::soundserver->getAudioSpec());
          glidersound->setMinRelVelocity(flMinRelV);
          glidersound->setMaxRelVelocity(flMaxRelV);
          glidersound->setMaxDistanceFeet(flMaxDist);
          sample = glidersound;
        }
        else
        {
          sample = new T_EngineSound(soundfile.c_str(), Global::soundserver->getAudioSpec());
        }
                
        sample->setType(sound_type);
        sample->setPitchFactor(dPitchFactor);
        sample->setMaxVolume(dMaxVolume);
        sample->setChannel(Global::soundserver->playSample((T_SoundSample*)sample));
        sound.push_back(sample);
      }
    }
  }
}

// todo: remove this method when transition to xml is done
void CRRCAirplaneLaRCSim::initSound(const char *airfile, double dPitchFactor)
{
  // assemble relative path
  std::string soundfile;
        
  soundfile = getParamAsString(airfile, "engine_sound");

  if (soundfile != "")
  {
    // Get full path (considering search paths). 
    soundfile = T_Config::getDataPath("sounds/" + soundfile);
  }
  
  // File ok? Use default otherwise.
  if (!T_Config::fileExists(soundfile))
    soundfile = T_Config::getDataPath("sounds/fan.wav");

  std::cout << "soundfile: " << soundfile << "\n";
  std::cout << "max_thrust: " << max_thrust << "\n";
  std::cout << "soundserver: " << Global::soundserver << "\n";
  
  // Only make noise if a sound file is available and if the model is able to provide thrust.
  if (soundfile != "" && Global::soundserver != (CRRCAudioServer*)0)
  {    
    long mode;    
    T_AirplaneSound *sample;
    if (!getParamAsLong(airfile, "engine_type", &mode))
    {
      // mostly gliders:
      mode = SOUND_TYPE_GLIDER;
    }    
    int sound_type = (int)mode;
    
    std::cout << "Using airplane sound " << soundfile << ", type " << sound_type << std::endl;
    
    if (mode == SOUND_TYPE_GLIDER)
    {
      sample = new T_GliderSound(soundfile.c_str(), Global::soundserver->getAudioSpec());
    }
    else
    {
      sample = new T_EngineSound(soundfile.c_str(), Global::soundserver->getAudioSpec());
    }
            
    sample->setType(sound_type);
    sample->setPitchFactor(dPitchFactor);
    sample->setChannel(Global::soundserver->playSample((T_SoundSample*)sample));
    sound.push_back(sample);
  }
  else
  {
    std::cout << "No engine sound. Enjoy the silence!" << std::endl;
  }
}

/*****************************************************************************/

#ifdef EXPERIMENTAL_STENCIL_SHADOW
int shadowPredrawCallback(ssgState*)
{
  glStencilMask(1);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_GREATER, 1, 1);
  glStencilOp(GL_KEEP, GL_REPLACE, GL_REPLACE);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  return 0;
}

int shadowPostdrawCallback(ssgState*)
{
  glStencilMask(0);
  glDisable(GL_STENCIL_TEST);
  return 0;
}
#endif


/** \brief Create a "shadow" instance of a model
 *
 *  This function clones a model and sets all associated
 *  states to a "shadowy" look. If the video context
 *  supports stencil buffering, the shadow will be
 *  translucent, else it will be pitch black.
 *
 *  \param ent The entity to be cloned
 */
void makeShadow(ssgEntity *ent)
{
  if (ent->isAKindOf(ssgTypeLeaf()))
  {
    ssgLeaf *leaf = (ssgLeaf*)ent;
    if (leaf->hasState())
    {
      ssgSimpleState *state = (ssgSimpleState*)leaf->getState();
      state->disable(GL_COLOR_MATERIAL);
      state->disable(GL_TEXTURE_2D);
      state->enable(GL_LIGHTING);
      state->enable(GL_BLEND);
      state->setShadeModel(GL_SMOOTH);
      state->setShininess(0.0f);
      state->setMaterial(GL_EMISSION, 0.0, 0.0, 0.0, 0.0);
      
      #ifdef EXPERIMENTAL_STENCIL_SHADOW
      if (vidbits.stencil)
      {
        state->setMaterial(GL_AMBIENT, 0.0, 0.0, 0.0, 0.1);
        state->setMaterial(GL_DIFFUSE, 0.0, 0.0, 0.0, 0.1);
        state->setMaterial(GL_SPECULAR, 0.0, 0.0, 0.0, 0.1);
        state->setStateCallback(SSG_CALLBACK_PREDRAW, shadowPredrawCallback);
        state->setStateCallback(SSG_CALLBACK_POSTDRAW, shadowPostdrawCallback);
      }
      else
      #endif
      {
        state->setMaterial(GL_AMBIENT, 0.0, 0.0, 0.0, 1.0);
        state->setMaterial(GL_DIFFUSE, 0.0, 0.0, 0.0, 1.0);
        state->setMaterial(GL_SPECULAR, 0.0, 0.0, 0.0, 1.0);
        state->setStateCallback(SSG_CALLBACK_PREDRAW, NULL);
        state->setStateCallback(SSG_CALLBACK_POSTDRAW, NULL);
      }
    }
  }
  else if (ent->isAKindOf(ssgTypeBranch()))
  {
    ssgBranch *branch = (ssgBranch*)ent;
    int kids = branch->getNumKids();
    for (int i = 0; i < kids; i++)
    {
      makeShadow(branch->getKid(i));
    }
  }
}


/** \brief The constructor.
 *
 *  Loads an airplane model from an xml description. The 3D model
 *  will be added to the specified scenegraph.
 *
 *  \param  xml   XML model description file
 *  \param  graph Pointer to the scenegraph which shall render the model
 */
CRRCAirplaneLaRCSimSSG::CRRCAirplaneLaRCSimSSG(SimpleXMLTransfer* xml, ssgBranch *graph)
  : CRRCAirplaneLaRCSim(xml), initial_trans(NULL), 
    model_trans(NULL), model(NULL),
    shadow(NULL), shadow_trans(NULL)
{
  printf("CRRCAirplaneLaRCSimSSG(xml, branch)\n");
  
  std::string s;      
  s = XMLModelFile::getGraphics(xml)->getString("model");
        
  // plib automatically loads the texture file, but it does not know which directory to use.
  {
    // where is the object file?
    std::string    of  = T_Config::getDataPath("objects/" + s);
    // compile and set relative texture path
    std::string    tp  = of.substr(0, of.length()-s.length()-1-7) + "textures";    
    ssgTexturePath(tp.c_str());
    // load model
    model = ssgLoad(of.c_str());
  }

  if (model != NULL)
  {
    // Offset of center of gravity
    CRRCMath::Vector3  pCG;         
    pCG = CRRCMath::Vector3(0, 0, 0);
    if (xml->indexOfChild("CG") >= 0)
    {
      SimpleXMLTransfer* i;
      i = xml->getChild("CG");
      pCG.r[0] = i->attributeAsDouble("x", 0);
      pCG.r[1] = i->attributeAsDouble("y", 0);
      pCG.r[2] = i->attributeAsDouble("z", 0);
      
      if (i->attributeAsInt("units") == 1)
        pCG *= M_TO_FT;
    }
    
    // transform model from SSG coordinates to CRRCsim coordinates
    initial_trans = new ssgTransform();
    model_trans = new ssgTransform();
    graph->addKid(model_trans);
    model_trans->addKid(initial_trans);
    initial_trans->addKid(model);
    
    sgMat4 it = {  {1.0,  0.0,  0.0,  0},
                   {0.0,  0.0, -1.0,  0},
                   {0.0,  1.0,  0.0,  0},
                   {pCG.r[1],  pCG.r[2],  -pCG.r[0],  1.0} };
    
    initial_trans->setTransform(it);

    // add a simple shadow
    shadow = (ssgEntity*)initial_trans->clone(SSG_CLONE_RECURSIVE | SSG_CLONE_GEOMETRY | SSG_CLONE_STATE);
    makeShadow(shadow);
    shadow_trans = new ssgTransform();
    graph->addKid(shadow_trans);
    shadow_trans->addKid(shadow);
    
    // add animations ("real" model only, without shadow)
    initAnimations(xml, model, &Global::inputs);
  }
  else
  {
    std::string msg = "Unable to open airplane model file \"";
    msg += s;
    msg += "\"\nspecified in \"";
    msg += xml->getSourceDescr();
    msg += "\"";

    throw std::runtime_error(msg);
  }

}

/** \brief The constructor.
 *
 *  Loads an airplane model from an .air file. The 3D model
 *  will be added to the specified scenegraph.
 *  \todo remove this method when transition to xml is done
 */
CRRCAirplaneLaRCSimSSG::CRRCAirplaneLaRCSimSSG(const char *filename, ssgBranch *graph)
  : CRRCAirplaneLaRCSim(filename), initial_trans(NULL), 
    model_trans(NULL), model(NULL),
    shadow(NULL), shadow_trans(NULL)
{
  printf("CRRCAirplaneLaRCSimSSG(\"%s\", branch)\n", filename);
  
  std::string s;
  
  s = getParamAsString(filename, "model");
  
  if (s == "")
  {
    std::string msg = "No \"model\" statement in airplane specification file \"";
    msg += filename;
    msg += "\"";

    throw std::runtime_error(msg);
  }

  
  // plib automatically loads the texture file, but it does not know which directory to use.
  {
    // where is the object file?
    std::string    of  = T_Config::getDataPath("objects/" + s);
    // compile and set relative texture path
    std::string    tp  = of.substr(0, of.length()-s.length()-1-7) + "textures";    
    ssgTexturePath(tp.c_str());
    // load model
    model = ssgLoad(of.c_str());
  }

  if (model != NULL)
  {
    initial_trans = new ssgTransform();
    model_trans = new ssgTransform();
    graph->addKid(model_trans);
    model_trans->addKid(initial_trans);
    initial_trans->addKid(model);
    
    sgMat4 it = {  {1.0,  0.0,  0.0,  0.0},
                   {0.0,  0.0, -1.0,  0.0},
                   {0.0,  1.0,  0.0,  0.0},
                   {0.0,  0.0,  0.0,  1.0}  };
    
    initial_trans->setTransform(it);

    shadow = (ssgEntity*)initial_trans->clone(SSG_CLONE_RECURSIVE | SSG_CLONE_GEOMETRY | SSG_CLONE_STATE);
    makeShadow(shadow);
    shadow_trans = new ssgTransform();
    graph->addKid(shadow_trans);
    shadow_trans->addKid(shadow);
    
  }
  else
  {
    std::string msg = "Unable to open airplane model file \"";
    msg += s;
    msg += "\"\nspecified in \"";
    msg += filename;
    msg += "\"";

    throw std::runtime_error(msg);
  }

}

CRRCAirplaneLaRCSimSSG::~CRRCAirplaneLaRCSimSSG()
{
  model_trans->getParent(0)->removeAllKids();
}

#ifdef SOME_TESTS
inline void makeEulerZXZRotMat4(sgMat4 mat, double phi, double theta, double psi)
{
  float cpsi = (float)cos(psi);
  float cphi = (float)cos(phi);
  float cthe = (float)cos(theta);
  float spsi = (float)sin(psi);
  float sphi = (float)sin(phi);
  float sthe = (float)sin(theta);
  
  mat[0][0] = cpsi * cphi - cthe * sphi * spsi;
  mat[0][1] = -1 * spsi * cphi - cthe * sphi * cpsi;
  mat[0][2] = sthe * sphi;
  mat[0][3] = SG_ZERO;
  
  mat[1][0] = cpsi * sphi + cthe * cphi * spsi;
  mat[1][1] = -1 * spsi * sphi + cthe * cphi * cpsi;
  mat[1][2] = -1 * sthe * cphi;
  mat[1][3] = SG_ZERO;
  
  mat[2][0] = spsi * sthe;
  mat[2][1] = cpsi * sthe;
  mat[2][2] = cthe;
  mat[2][3] = SG_ZERO;
  
  mat[3][0] = SG_ZERO;
  mat[3][1] = SG_ZERO;
  mat[3][2] = SG_ZERO;
  mat[3][3] = SG_ONE;
}

inline void makeEulerZYZRotMat4(sgMat4 mat, double phi, double theta, double psi)
{
  float cpsi = (float)cos(psi);
  float cphi = (float)cos(phi);
  float cthe = (float)cos(theta);
  float spsi = (float)sin(psi);
  float sphi = (float)sin(phi);
  float sthe = (float)sin(theta);
  
  mat[0][0] = -1 * spsi * sphi + cthe * cphi * cpsi;
  mat[0][1] = -1 * cpsi * sphi - cthe * cphi * spsi;
  mat[0][2] = sthe * cphi;
  mat[0][3] = SG_ZERO;
  
  mat[1][0] = spsi * cphi + cthe * sphi * cpsi;
  mat[1][1] = cpsi * cphi - cthe * sphi * spsi;
  mat[1][2] = sthe * sphi;
  mat[1][3] = SG_ZERO;
  
  mat[2][0] = -1 * cpsi * sthe;
  mat[2][1] = spsi * sthe;
  mat[2][2] = cthe;
  mat[2][3] = SG_ZERO;
  
  mat[3][0] = SG_ZERO;
  mat[3][1] = SG_ZERO;
  mat[3][2] = SG_ZERO;
  mat[3][3] = SG_ONE;
}
#endif


/** \brief Create a rotation matrix
 *
 *  This function creates a rotation matrix from the original
 *  OpenGL glRotatef commands in CRRCAirplaneLaRCSim::draw().
 *
 *  \param m The matrix to be rotated
 *  \param phi Euler angle phi
 *  \param theta Euler angle theta
 *  \param psi Euler angle psi
 */
inline void makeOGLRotMat4(sgMat4 m, double phi, double theta, double psi)
{
#if COOR_NEW == 1
  sgMat4 temp;
  sgVec3 rvec;
  
  //~ sgSetVec3(rvec, 0.0, 1.0, 0.0);
  //~ sgMakeRotMat4(temp, 90.0, rvec);
  //~ sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, 0.0, 1.0, 0.0);
  sgMakeRotMat4(temp, 180.0f - (float)psi * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, -1.0, 0.0, 0.0);
  sgMakeRotMat4(temp, (float)theta * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, 0.0, 0.0, 1.0);
  sgMakeRotMat4(temp, (float)phi * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);
  
#else

  sgMat4 temp;
  sgVec3 rvec;
  
  sgSetVec3(rvec, 0.0, 1.0, 0.0);
  sgMakeRotMat4(temp, 90.0, rvec);
  sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, 0.0, -1.0, 0.0);
  sgMakeRotMat4(temp, (float)psi * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, -1.0, 0.0, 0.0);
  sgMakeRotMat4(temp, -1 * (float)theta * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);
  
  sgSetVec3(rvec, 0.0, 0.0, 1.0);
  sgMakeRotMat4(temp, (float)phi * SG_RADIANS_TO_DEGREES, rvec);
  sgPreMultMat4(m, temp);

#endif
}


/** \brief Draw the airplane
 *
 *  This method actually does not draw anything. It only
 *  updates the aircraft's transformation node in the
 *  scenegraph with the aircraft's current position and
 *  orientation. The actual drawing is done by the global
 *  ssgCullAndDraw() call.
 *
 *  \param airplane pointer to the airplane's FDM object
 */
void CRRCAirplaneLaRCSimSSG::draw(CRRC_AirplaneSim* airplane)
{
  CRRCMath::Vector3 pos = airplane->getPos();
  
  sgMat4 m;
  sgMakeIdentMat4(m);

#if COOR_NEW == 1
  m[3][0] = pos.r[1];
  m[3][1] = -1 * pos.r[2];
  m[3][2] = -1 * pos.r[0];
#else
  m[3][0] = pos.r[0];
  m[3][1] = -1 * pos.r[2];
  m[3][2] = pos.r[1];
#endif
  makeOGLRotMat4(m, airplane->getPhi(), airplane->getTheta(), airplane->getPsi());
  model_trans->setTransform(m);
  
}


/** \brief Calculate pitch and roll from a given normal vector.
 *
 *  Stolen from the tuxkart sources.
 *
 *  \param hpr  destination vector
 *  \param nrm  normal vector
 */
inline void pr_from_normal ( sgVec3 hpr, sgVec3 nrm )
{
  float sy = sin ( -hpr [ 0 ] * SG_DEGREES_TO_RADIANS ) ;
  float cy = cos ( -hpr [ 0 ] * SG_DEGREES_TO_RADIANS ) ;

  hpr[2] =  SG_RADIANS_TO_DEGREES * atan2 ( nrm[0] * cy - nrm[1] * sy, nrm[2] ) ;
  hpr[1] = -SG_RADIANS_TO_DEGREES * atan2 ( nrm[1] * cy + nrm[0] * sy, nrm[2] ) ;
}


/** \brief Draw the airplane shadow
 *
 *  This method actually does not draw anything. It only
 *  updates the aircraft shadow's transformation node in the
 *  scenegraph with the aircraft shadow's current position and
 *  orientation. The actual drawing is done by the global
 *  ssgCullAndDraw() call.
 *
 *  \todo calculate proper shadow rotation
 *  \todo adjust shadow location to match real terrain height
 *
 *  \param airplane pointer to the airplane's FDM object
 */
void CRRCAirplaneLaRCSimSSG::draw_shadow(CRRC_AirplaneSim* airplane,
                 float             shadow_matrix[4][4])
{
  CRRCMath::Vector3 pos = airplane->getPos();

  sgMat4 m;

  model_trans->getTransform(m);
  
  sgPostMultMat4(m, shadow_matrix);
//jwtodoshadow  m[3][1] += 0.1;
  
  shadow_trans->setTransform(m);
}
