/*****************************************************************************
Title: CRRCsim, the Charles River Radio Control Club Flight Simulator Project
Authors:
Jan Kansky:  Programming
Mark Drela:  Aerodynamics
Jan Reucker
Jens Wilhelm Wulf
 
Purpose:  The idea for CRRCsim is to take away any last excuse you might
have for not using a flight simulator to keep your thumbs certified for RC
flying.   It's free, it works, enjoy.  This is an open source project, so if
you don't like the way something works, help us fix it!
 
Thanks:
  Bruce Jackson for the LaRCsim framework.
  Flight Gear project for the sky sphere.
 
Contacts:
If you'd like to help with CRRCSIM, then send me an email!
    email                : kansky@ll.mit.edu
 
File:   crrc_loadair.h
        Routines dealing with .air file loading
 
*****************************************************************************/

#ifndef CRRC_LOADAIR_H
# define CRRC_LOADAIR_H

# include <string>
# include <vector>
# include <stdexcept>

# include <plib/ssg.h>

# include "global.h"
# include "include_gl.h"
# include "fdm.h"
# include "crrc_sound.h"
# include "SimpleXMLTransfer.h"

extern GLUquadricObj* quadric;

typedef struct
{
   char *extrusion_name;
   gleDouble up[3];
   int num_points_on_path;
   GLfloat translate_x,translate_y,translate_z;
   GLfloat rotate_x,rotate_y,rotate_z;
   gleDouble **point_array;
   gleAffine *scaling;
   GLfloat **colors;
   int num_contour_elements;
   gleDouble **contour;
   gleDouble **contour_normal_vectors;
   void *next_extrusion;
} extrusion;

typedef struct
{
   char *cylinder_name;
   GLfloat translate_x,translate_y,translate_z;
   GLfloat rotate_x,rotate_y,rotate_z;
   GLfloat radius_1,radius_2,length;
   GLfloat radial_sampling,longitudinal_sampling;
   GLfloat colors[4];
   void *next_cylinder;
} cylinder;

typedef struct
{
   char *sphere_name;
   GLfloat translate_x,translate_y,translate_z;
   GLfloat rotate_x,rotate_y,rotate_z;
   GLfloat radius;
   GLfloat slices,stacks;
   GLfloat colors[4];
   void *next_sphere;
} sphere;

typedef struct
{
   char *triangle_name;
   unsigned long int num_triangles;
   GLfloat translate_x,translate_y,translate_z;
   GLfloat rotate_x,rotate_y,rotate_z;
   GLfloat colors[4];
   GLfloat **vertices;
   GLfloat **normals;
   void *next_triangle;
} triangle;


/** \brief Read a long parameter from an .air file.
 *
 *  Reads a parameter from an .air file and places the parameter's
 *  value in the variable pointed to by <code>value</code>.
 *
 *  If the parameter name is found in the file, the returned
 *  status will always be <code>true</code>, even if no valid
 *  value was found behind the parameter name. In this case,
 *  the content of <code>value</code> will be undefined.
 *
 *  \param datafile file name
 *  \param param    parameter name
 *  \param value    parameter value as read from the file
 *  \return true if the parameter was found in the file, false otherwise
 */
bool getParamAsLong(const char *datafile, const char *param, long *value);

/** \brief Read a string parameter from an .air file.
 *
 *  \param datafile file name
 *  \param param    parameter name
 *  \return parameter value as a std::string
 */
std::string getParamAsString(const char *datafile, const char *param);


/** \brief A very basic airplane description class.
 *
 *  This is a first attempt to organize all airplane-related
 *  stuff into a class. At this time it's nothing more than
 *  a wrapper for the existing functions.
 * 
 *  Update: this class should contain everything but the simulation of
 *  the EOM (all the physics stuff). So this is about visualization.
 */
class CRRCAirplane
{
  public:
   
   /** \brief The constructor.
    *
    *  
    */
   CRRCAirplane();
   
   /** \brief The destructor.
    *
    *  Clean up.
    */
   virtual ~CRRCAirplane();
   
   /** \brief Draw the precompiled airplane shadow.
    *
    *  Applies some material properties and calls the
    *  shadow's display list.
    */
   virtual void draw_shadow(CRRC_AirplaneSim* airplane,
                            float             shadow_matrix[4][4]) = 0;
   
   /** \brief Draw the airplane
    *
    *
    */
   virtual void draw(CRRC_AirplaneSim* airplane) = 0;

   virtual int  getNumSounds()  {return (sound.size());};
   //~ virtual int  getSoundType(int soundnum)  {return sound[soundnum]->type;};
   //~ virtual double getSoundPitchFactor(int soundnum) { return(sound[soundnum]->dPitchFactor); };
  
  protected:
   std::vector<T_AirplaneSound*> sound;
};


/**
 * Reads geometric data directly from old-style .air-files.
 */
class CRRCAirplaneLaRCSim : public CRRCAirplane
{
  public:
   CRRCAirplaneLaRCSim();
   CRRCAirplaneLaRCSim(const char* filename);
   CRRCAirplaneLaRCSim(SimpleXMLTransfer* xml);
   virtual ~CRRCAirplaneLaRCSim();

   virtual void draw(CRRC_AirplaneSim* airplane);
   virtual void draw_shadow(CRRC_AirplaneSim* airplane,
                            float             shadow_matrix[4][4]);

  protected:
   
  /** \brief Initialize the airplane's sound.
    *
    *  Reads all sound related parameters from an .air file.
    *  \todo Make this method search all possible paths on Linux!
    */
   virtual void  initSound(const char *airfile, double dPitchFactor);
   
  /** \brief Initialize the airplane's sound.
    *
    *  Reads all sound related parameters from an xml description.
    *  \todo Make this method search all possible paths on Linux!
    */
   virtual void  initSound(SimpleXMLTransfer* xml);

   float max_thrust;
  
  private:
    
   void dealloc_plane_model();

   /** \brief Compile airplane display list
    *
    *  This function creates an OpenGL display list from
    *  the airplane's 3D model data. All extrusions and
    *  calculations are only done once at startup. After
    *  that all GL calls are done with static data by just
    *  calling the list.
    */
   void compile_airplane();

   /** \brief Compile airplane shadow display list
    *
    *  This function creates an OpenGL display list from
    *  the airplane's 3D model data, representing the
    *  airplane's ground shadow. All extrusions and
    *  calculations are only done once at startup. After
    *  that all GL calls are done with static data by just
    *  calling the list.
    */
   void compile_airplane_shadow();   
   
  private:
   gleDouble **contour;       // The 2D airfoil contour
   gleDouble **point_array;   // Those 3D spanwise rib leading edge locations
   gleAffine *wing_scaling;   // Scaling of airfoil pattern for each rib
   float **wing_colors;       // Color at each rib location
      
   extrusion* extrusions;
   cylinder*  cylinders;
   sphere*    spheres;
   triangle*  triangles;

   // OpenGL-lists for plane and shadow
   GLuint airplane_list;
   GLuint airplane_shadow_list;

   GLUquadricObj *quadric;
};


/** \brief Display airplane model using SSG
 *
 *  This airplane class also uses old-style .air-files,
 *  but does not use any of the geometry descriptions in
 *  the file. Instead it loads the model specified by
 *  the "model" parameter in the file.
 */
class CRRCAirplaneLaRCSimSSG : public CRRCAirplaneLaRCSim
{
  public:
    CRRCAirplaneLaRCSimSSG(const char*        filename, ssgBranch* graph);
    CRRCAirplaneLaRCSimSSG(SimpleXMLTransfer* xml,      ssgBranch* graph);
   
    virtual ~CRRCAirplaneLaRCSimSSG();
    
    virtual void draw(CRRC_AirplaneSim* airplane);
    virtual void draw_shadow(CRRC_AirplaneSim* airplane,
                            float             shadow_matrix[4][4]);

  
  private:
    ssgTransform  *initial_trans;
    ssgTransform  *model_trans;
    ssgEntity     *model;
    ssgEntity     *shadow;
    ssgTransform  *shadow_trans;
};


#endif // CRRC_LOADAIR_H
