// -*- mode: c; mode: fold -*-
//
#include "fdm_002.h"

#include <math.h>
#include <iostream>

#include "../global.h"
#include "../defines.h"
#include "../ls_constants.h"
#include "../crrc_scenery.h"
#include "../crrc_main.h" // because of "scenery"
#include "../windfield.h"
#include "../lib_conversions.h"
#include "../xmlmodelfile.h"


/**
 * Value for non-SI (original LaRCSim): 0.0023756 slug/(ft)^3
 */
#define DENSITY 0.0023756

// 0
// 1
// 2
// 3: simple test of drawing transformation
#define EOM_TEST 0

void CRRC_AirplaneSim_002::initAirplaneState(double dRelVel,
                                             double dTheta,
                                             double dPsi,
                                             double X,
                                             double Y,
                                             double Z) /*{{{*/
{
  float       flVelocity = dRelVel * getTrimmedFlightVelocity();
  
  // About inertia:
  //    |  I_xx  -I_xy  -I_xz |
  //    | -I_xy   I_yy  -I_yz |
  //    | -I_xz  -I_yz   I_zz |
  // I_xx, I_yy, I_zz are needed in any case.
  // If the plane is symmetric about its y-axis, I_yz=I_xy=0.
  //    |  I_xx   0     -I_xz |
  //    |  0      I_yy   0    |
  //    | -I_xz   0      I_zz |
  // 
  // What about the sign of I_xz?
  //    I_xz = integral x * z dm
  // 
  // Four possibilities:
  //   1. positive I_xz  
  //      1.1 in front of cg (x>0) and below cg (z>0)
  //      1.2 behind cg (x<0) and over cg (z<0)
  //   2. negative I_xz
  //      2.1 behind cg (x<0) and below cg (z>0)
  //      2.2 in front of cg (x>0) and over cg (z<0)
  //     
  //                    |
  //                    |
  //        negative    |   positive
  //                    |
  //                    |
  //   x <--------------+-----------------
  //                    |
  //                    |
  //        positive    |   negative
  //                    |
  //                    |
  //                    v z
  // 
  // It's easy to see that the vertical fin of the rudder gives a positive I_xz.
  // Usually the nose of the fuselage is heavy/long and below the cg, which leads to 
  // a positive I_xz, too.
  // But one is free to choose the direction of the body axes (let the x-axis point a little 
  // higher or lower from the cg) and thereby change I_xz.
  
  // jwtodo: Anstellwinkel gescheit einbauen!
  eom = EOM_6DOF(CRRCMath::Vector3(X, Y, Z),
                 CRRCMath::Vector3(0, dTheta, dPsi),
                 CRRCMath::Vector3(flVelocity, 0, 0),
                 Mass,
                 CRRCMath::Matrix33( I_xx,    0,    -I_xz,
                                     0,       I_yy,  0,
                                    -I_xz,    0,     I_zz));

#if (EOM_TEST != 0)
  eom.setGravity(0);
#else
  // we are not using SI-units:
  eom.setGravity(9.81*METERS_TO_FT);
#endif
  
  v_V_local_airmass = CRRCMath::Vector3(); // local velocity of steady airmass   [ft/s]
  
  // jwtodo: seems this can be used to simulate wind gusts
  v_V_gust_local  = CRRCMath::Vector3();
  
  CD_stall        = 0.05;   // drag coeff. during stalling    []
  
  m_V_atmo_rwy = CRRCMath::Matrix33();

  power->reset();
  
  aero_init();
}
/*}}}*/

void CRRC_AirplaneSim_002::update(TSimInputs* inputs,
                                  double      dt,
                                  int         multiloop) /*{{{*/
{
  double V_north_xp,V_east_xp,V_down_xp;   // temp vars for calculating wind grad.
  double V_north_yp,V_east_yp,V_down_yp;
  double V_north_zp,V_east_zp,V_down_zp;
  double V_north_xm,V_east_xm,V_down_xm;
  double V_north_ym,V_east_ym,V_down_ym;
  double V_north_zm,V_east_zm,V_down_zm;

  /**
   * Using a length of about roughly one half of the aircrafts
   * size to calculate wind gradients. 0.1 foot had been used before,
   * which leads to very high or zero gradients.
   */
  double delta_space = getAircraftSize()/2;
  
  // Anpassung an LaRCSim: dort ist H=-Z
  CRRCMath::Vector3 v_P_CG_Rwy = eom.pos.val;
  
  int   nAircraftOutsideWindfieldSim =
    calculate_wind(v_P_CG_Rwy.r[0]+delta_space, v_P_CG_Rwy.r[1],             v_P_CG_Rwy.r[2],
                   V_north_xp,                  V_east_xp,                   V_down_xp) |
    calculate_wind(v_P_CG_Rwy.r[0],             v_P_CG_Rwy.r[1]+delta_space, v_P_CG_Rwy.r[2],
                   V_north_yp,                  V_east_yp,                   V_down_yp) |
    calculate_wind(v_P_CG_Rwy.r[0],             v_P_CG_Rwy.r[1],             v_P_CG_Rwy.r[2]+delta_space,
                   V_north_zp,                  V_east_zp,                   V_down_zp) |
    calculate_wind(v_P_CG_Rwy.r[0]-delta_space, v_P_CG_Rwy.r[1],             v_P_CG_Rwy.r[2],
                   V_north_xm,                  V_east_xm,                   V_down_xm) |
    calculate_wind(v_P_CG_Rwy.r[0],             v_P_CG_Rwy.r[1]-delta_space, v_P_CG_Rwy.r[2],
                   V_north_ym,                  V_east_ym,                   V_down_ym) |
    calculate_wind(v_P_CG_Rwy.r[0],             v_P_CG_Rwy.r[1],             v_P_CG_Rwy.r[2]-delta_space,
                   V_north_zm,                  V_east_zm,                   V_down_zm) |
    calculate_wind(v_P_CG_Rwy.r[0],             v_P_CG_Rwy.r[1],             v_P_CG_Rwy.r[2],
                   v_V_local_airmass.r[0],      v_V_local_airmass.r[1],      v_V_local_airmass.r[2]);  
  
  if (nAircraftOutsideWindfieldSim)
  {
    // todo: some error message?
  }
  
  // Gradients are calculated from symmetric pairs to get symmetric behaviour.
  m_V_atmo_rwy.v[0][0] = (V_north_xp - V_north_xm)/(2*delta_space);
  m_V_atmo_rwy.v[0][1] = (V_north_yp - V_north_ym)/(2*delta_space);
  m_V_atmo_rwy.v[0][2] = (V_north_zp - V_north_zm)/(2*delta_space);
  m_V_atmo_rwy.v[1][0] = (V_east_xp  - V_east_xm) /(2*delta_space);
  m_V_atmo_rwy.v[1][1] = (V_east_yp  - V_east_ym) /(2*delta_space);
  m_V_atmo_rwy.v[1][2] = (V_east_zp  - V_east_zm) /(2*delta_space);
  m_V_atmo_rwy.v[2][0] = (V_down_xp  - V_down_xm) /(2*delta_space);
  m_V_atmo_rwy.v[2][1] = (V_down_yp  - V_down_ym) /(2*delta_space);
  m_V_atmo_rwy.v[2][2] = (V_down_zp  - V_down_zm) /(2*delta_space);

#if (EOM_TEST == 2)
  /*
  switch (nStep)
  {
   case 0:
    if (inputs->elevator < 0)
      nStep = 1;
    multiloop = 0;
    break;
   case 2:
    if (inputs->elevator > 0)
      nStep = 3;
    multiloop = 0;
    break;
   case 4:
    nStep     = 0;
    multiloop = 0;
   default:
    break;
  }
   */
#endif
  
  for (int n=0; n<multiloop; n++)
  {        
    logNewline();
    
#if FDM_LOG_POS != 0
    logVal(eom.pos.val);
    logVal(getPhi());    
    logVal(getTheta());
    logVal(getPsi());    
#endif    
#if FDM_LOG_WIND_IN != 0
    logVal(v_V_local_airmass);
    logVal(m_V_atmo_rwy);
#endif    
    
#if (EOM_TEST != 2)
    aero( dt, inputs);
#endif

#if FDM_LOG_AERO_OUT != 0
    logVal(v_F_aero);
    logVal(v_M_aero);
#endif
    
    {
      v_F_engine = CRRCMath::Vector3();
      v_M_engine = CRRCMath::Vector3();

#if (EOM_TEST != 2)
      power->step(dt, inputs, v_V_body.r[0]*FT_TO_METERS, &v_F_engine, &v_M_engine);
#endif
      
      // Convert SI to that other buggy system.
      v_F_engine *= N_TO_LBF;
      v_M_engine *= NM_TO_LBFFT;
    }
    gear(inputs);

#if (EOM_TEST == 2)
    CRRCMath::Vector3 v_F, v_M_cg;
    
    switch ((nLogCnt>>5) % 6)
    {
     case 0:
      v_F    = CRRCMath::Vector3(1, 0, 0);
      v_M_cg = CRRCMath::Vector3(0, 0, 0);
      break;
     case 1:
      v_F    = CRRCMath::Vector3(0, 0, 0);
      v_M_cg = CRRCMath::Vector3(0, 2, 0);
      break;
     case 2:
      v_F    = CRRCMath::Vector3(0, 0, 3);
      v_M_cg = CRRCMath::Vector3(0, 0, 0);
      break;
     case 3:
      v_F    = CRRCMath::Vector3(0, 0, 0);
      v_M_cg = CRRCMath::Vector3(1, 0, 0);
      break;
     case 4:
      v_F    = CRRCMath::Vector3(0, 2, 0);
      v_M_cg = CRRCMath::Vector3(0, 0, 0);
      break;
     case 5:
      v_F    = CRRCMath::Vector3(0, 0, 0);
      v_M_cg = CRRCMath::Vector3(0, 0, 3);
      break;
    }
  
    {
      double d = sin(M_PI*(nLogCnt&0x1F)/32.0);
      
      v_F    *= d*d;
      v_M_cg *= d*d;
    }
    
    if ((nLogCnt&0x1F) == 0x1F)
    {
      nStep++;
      std::cout << "\n";
    }
    
    if (  ((nLogCnt>>5)/6) % 2)
    {
      v_F    *= -1;
      v_M_cg *= -1;
    }
    
//    v_F *= 0;

    v_F.print("F=", ", ");
    v_M_cg.print("M=", "\n");
    
    eom.step(dt, v_F, v_M_cg);
    eom.conv.updateEuler();
#else
    eom.step(dt, 
             v_F_aero + v_F_engine + v_F_gear,
             v_M_aero + v_M_engine + v_M_gear);
    
    // Update the position they are interested in outside of this fdm. I have to
    // do it inside of the loop because gear() also needs it.
    eom.conv.updateEuler();
#endif
    
//    eom.conv.convTest1();
  }
     
  /*
  static int cnt  = 0;
  static int posi = 0;
  bool fChange = false;
   */
  
  /*
  posi = 0;
  if (inputs->aileron > 0)
    posi |= 1;
  if (inputs->elevator > 0)
    posi |= 2;
  fChange = true;
   */
  
  /*
  cnt += multiloop;
  if (cnt > 6000)
  {
    cnt -= 6000;
    posi++;
    posi %= 4;
    fChange = true;
  }
   */
  
  /*
  switch (posi)
  {
   case 0:
    pos->d_x = 20;
    pos->d_y = 20;
    pos->d_z = -5;
    break;
   case 1:
    pos->d_x = -20;
    pos->d_y = 20;
    pos->d_z = -6;
    break;
   case 3:
    pos->d_x = -20;
    pos->d_y = -20;
    pos->d_z = -3;
    break;
   default:
    pos->d_x = 20;
    pos->d_y = -20;
    pos->d_z = -4;
    break;
  }
   */

  /*
  pos->d_x = inputs->elevator*30;
  pos->d_y = inputs->aileron*30;
  pos->d_z = -0.5*sqrt(pos->d_x*pos->d_x + pos->d_y*pos->d_y);
  
  eom.conv.euler.r[0] = (M_PI/2) -1.0 *atan2(pos->d_x, pos->d_y);
  eom.conv.euler.r[1] = (3.0*M_PI/4.0); // (M_PI/2) -1.0 *atan2(pos->d_x, pos->d_y);
  eom.conv.euler.r[2] = (M_PI/2); // (M_PI/2) -1.0 *atan2(pos->d_x, pos->d_y);
    */
  
  // fChange = true;
  
  /*
  if (fChange)
  {
    std::cout << pos->d_x << " " << pos->d_y << " " << pos->d_z << " "
      << eom.conv.euler.r[0] << " " << eom.conv.euler.r[1] << " " << eom.conv.euler.r[2] << "\n";
  }
  */
  
  
#if (EOM_TEST == 3)
  eom.pos.val      = CRRCMath::Vector3(10, 20, -5);
  eom.conv.euler   = CRRCMath::Vector3(0.25*PI, 0.25*PI, 0.25*PI);
#endif
  
}
/*}}}*/

CRRC_AirplaneSim_002::CRRC_AirplaneSim_002(const char* filename) : CRRC_AirplaneSim("fdm_002.dat") /*{{{*/
{
  FILE *airplane_data;
  char parameter_name[256]; //  Name of parameter
  char trailing_comment[256];
  double parameter_value;
  bool fSectionParsingDone;
  int loop;
  float max_thrust = 0.0;          // Maximum thrust from engine.   

  /**
   * Tracks wingspan [m]
   */
  double span = 0;
  span_ft = 0.0;

  power = (Power::Power*)0;
  
  C_ref = 0;
  B_ref = 0;
  S_ref = 0;
  U_ref = 0;
  
  Alpha_0 = 0;
  
  Cm_0 = 0;
  CL_0 = 0;
  CL_max = 0;
  CL_min = 0;
  CD_prof = 0;
  
  Uexp_CD = 0;
  
  CL_a = 0;
  Cm_a = 0;
  CY_b = 0;
  Cl_b = 0;
  Cn_b = 0;
  
  CL_q = 0;
  Cm_q = 0;
  CY_p = 0;
  Cl_p = 0;
  Cn_p = 0;
  CY_r = 0;
  Cl_r = 0;
  Cn_r = 0;
  
  CL_de = 0;
  Cm_de = 0;
  CY_dr = 0;
  Cl_dr = 0;
  Cn_dr = 0;
  CY_da = 0;
  Cl_da = 0;
  Cn_da = 0;
  
  eta_loc = 0;
  CG_arm = 0;
  CL_drop = 0;
  CD_stall = 0;
  
  span_eff=1.0; // Effective span  0.95 for most planes, 0.85 flying wing
  CL_CD0=0.0; // 0.30 for 7037, 0.15 MH32, 0.0 RG15, AGxx, power
  CD_CLsq=0.0; // 0.01 composites, 0.015 saggy ships, 0.02 beat up ship
  CD_AIsq=0.0; // 0.01/(max_aileron)^2
  CD_ELsq=0.0; // 0.01/(max_elevator)^2 for Zagi otherwise 0

  Mass = 0;
  I_xx = 0;
  I_yy = 0;
  I_zz = 0;
  I_xz = 0;
  
  airplane_data = fopen(filename, "r");
  if (airplane_data == NULL)
  {
    std::string msg = "Unable to open airplane specification file ";
    msg += filename;

    throw std::runtime_error(msg);
  }
  else
  {
    while (fscanf(airplane_data,"%s %lf %[^\n]",(char *)parameter_name,
                  &parameter_value,trailing_comment)!=EOF)
    {

      if (strcasecmp(parameter_name,"Mass")==0)
        Mass=parameter_value;
      else if (strcasecmp(parameter_name,"I_xx")==0)
        I_xx=parameter_value;
      else if (strcasecmp(parameter_name,"I_yy")==0)
        I_yy=parameter_value;
      else if (strcasecmp(parameter_name,"I_zz")==0)
        I_zz=parameter_value;
      else if (strcasecmp(parameter_name,"I_xz")==0)
        I_xz=parameter_value;
      else if (strcasecmp(parameter_name,"C_ref")==0)
        C_ref=parameter_value;
      else if (strcasecmp(parameter_name,"B_ref")==0)
        B_ref=parameter_value;
      else if (strcasecmp(parameter_name,"S_ref")==0)
        S_ref=parameter_value;
      else if (strcasecmp(parameter_name,"U_ref")==0)
        U_ref=parameter_value;
      else if (strcasecmp(parameter_name,"Alpha_0")==0)
        Alpha_0=parameter_value;
      else if (strcasecmp(parameter_name,"Cm_0")==0)
        Cm_0=parameter_value;
      else if (strcasecmp(parameter_name,"CL_0")==0)
        CL_0=parameter_value;
      else if (strcasecmp(parameter_name,"CL_max")==0)
        CL_max=parameter_value;
      else if (strcasecmp(parameter_name,"CL_min")==0)
        CL_min=parameter_value;
      else if (strcasecmp(parameter_name,"CD_prof")==0)
        CD_prof=parameter_value;
      else if (strcasecmp(parameter_name,"Uexp_CD")==0)
        Uexp_CD=parameter_value;
      else if (strcasecmp(parameter_name,"CL_a")==0)
        CL_a=parameter_value;
      else if (strcasecmp(parameter_name,"Cm_a")==0)
        Cm_a=parameter_value;
      else if (strcasecmp(parameter_name,"CY_b")==0)
        CY_b=parameter_value;
      else if (strcasecmp(parameter_name,"Cl_b")==0)
        Cl_b=parameter_value;
      else if (strcasecmp(parameter_name,"Cn_b")==0)
        Cn_b=parameter_value;
      else if (strcasecmp(parameter_name,"CL_q")==0)
        CL_q=parameter_value;
      else if (strcasecmp(parameter_name,"Cm_q")==0)
        Cm_q=parameter_value;
      else if (strcasecmp(parameter_name,"CY_p")==0)
        CY_p=parameter_value;
      else if (strcasecmp(parameter_name,"Cl_p")==0)
        Cl_p=parameter_value;
      else if (strcasecmp(parameter_name,"Cn_p")==0)
        Cn_p=parameter_value;
      else if (strcasecmp(parameter_name,"CY_r")==0)
        CY_r=parameter_value;
      else if (strcasecmp(parameter_name,"Cl_r")==0)
        Cl_r=parameter_value;
      else if (strcasecmp(parameter_name,"Cn_r")==0)
        Cn_r=parameter_value;
      else if (strcasecmp(parameter_name,"CL_de")==0)
        CL_de=parameter_value;
      else if (strcasecmp(parameter_name,"Cm_de")==0)
        Cm_de=parameter_value;
      else if (strcasecmp(parameter_name,"CY_dr")==0)
        CY_dr=parameter_value;
      else if (strcasecmp(parameter_name,"Cl_dr")==0)
        Cl_dr=parameter_value;
      else if (strcasecmp(parameter_name,"Cn_dr")==0)
        Cn_dr=parameter_value;
      else if (strcasecmp(parameter_name,"CY_da")==0)
        CY_da=parameter_value;
      else if (strcasecmp(parameter_name,"Cl_da")==0)
        Cl_da=parameter_value;
      else if (strcasecmp(parameter_name,"Cn_da")==0)
        Cn_da=parameter_value;
      else if (strcasecmp(parameter_name,"eta_loc")==0)
        eta_loc=parameter_value;
      else if (strcasecmp(parameter_name,"span_eff")==0)
        span_eff=parameter_value;
      else if (strcasecmp(parameter_name,"CL_CD0")==0)
        CL_CD0=parameter_value;
      else if (strcasecmp(parameter_name,"CD_CLsq")==0)
        CD_CLsq=parameter_value;
      else if (strcasecmp(parameter_name,"CD_AIsq")==0)
        CD_AIsq=parameter_value;
      else if (strcasecmp(parameter_name,"CD_ELsq")==0)
        CD_ELsq=parameter_value;
      else if (strcasecmp(parameter_name,"CG_arm")==0)
        CG_arm=parameter_value;
      else if (strcasecmp(parameter_name,"CL_drop")==0)
        CL_drop=parameter_value;
      else if (strcasecmp(parameter_name,"CD_stall")==0)
        CD_stall=parameter_value;
      else if (strcasecmp(parameter_name,"max_thrust")==0)
        max_thrust=parameter_value;
      else if (strcasecmp(parameter_name,"gear")==0)
      {
        bool  done_parsing_gear;
        float p1,p2,p3;      // Temporary variables used to read in file
        float dist;
        int   num_wheels;
        Wheel wheel;

        printf("In gear\n");

        // let's assume that there is nothing below the CG:
        dZLow    = 0;

        // let's assume that there is nothing distant from the CG:
        dMaxSize = 0;

        num_wheels=(int)parameter_value;
        wheels.clear();
        for (loop=0; loop<num_wheels; loop++)
          wheels.push_back(wheel);
        
        done_parsing_gear=false;
        while(!done_parsing_gear)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          if (strcasecmp(parameter_name,"locations")==0)
          {
            for (loop=0;loop<num_wheels;loop++)
            {
              fscanf(airplane_data,"%f %f %f",&p1,&p2,&p3);
              
              wheels[loop].v_P = CRRCMath::Vector3((double)p1, (double)p2, (double)p3);
              
              // track wingspan
              if (span < (DATA)p2)
                span = (DATA)p2;
              // lowest point?
              if (dZLow < (DATA)p3)
                dZLow = (DATA)p3;
              // far away (Z distance is assumed to be low)?
              dist = p1*p1+p2*p2;
              if (dist > dMaxSize)
                dMaxSize = dist;
            }
            dMaxSize = sqrt(dMaxSize);
            span_ft  = 2 * span;
            span     = span_ft * FT_TO_METERS;
          }
          else if (strcasecmp(parameter_name,"springiness")==0)
          {
            for (loop=0;loop<num_wheels;loop++)
            {
              fscanf(airplane_data,"%f",&p1);
              wheels[loop].spring_constant = (DATA)p1;
            }
          }
          else if (strcasecmp(parameter_name,"damping")==0)
          {
            for (loop=0;loop<num_wheels;loop++)
            {
              fscanf(airplane_data,"%f",&p1);
              wheels[loop].spring_damping=(DATA)p1;
            }
          }
          else if (strcasecmp(parameter_name,"braking")==0)
          {
            for (loop=0;loop<num_wheels;loop++)
            {
              fscanf(airplane_data,"%f",&p1);
              wheels[loop].percent_brake=(DATA)p1;
            }
          }
          else if (strcasecmp(parameter_name,"steerability")==0)
          {
            for (loop=0;loop<num_wheels;loop++)
            {
              fscanf(airplane_data,"%f %f",&p1,&p2);
              wheels[loop].steering_mapping   = (int)p1;
              wheels[loop].steering_max_angle = p2;
              wheels[loop].caster_angle_rad   = 0;
            }
          }
          else if (strcasecmp(parameter_name,"end_gear")==0)
          {
            done_parsing_gear=true;
          }
        }
      }
      else if (strcasecmp(parameter_name,"triangles")==0)
      {
        fSectionParsingDone = false;
        while (!fSectionParsingDone)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          fSectionParsingDone = (strcasecmp(parameter_name,"end_triangles")==0);
        }
      }
      else if (strcasecmp(parameter_name,"sphere")==0)
      {
        fSectionParsingDone = false;
        while (!fSectionParsingDone)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          fSectionParsingDone = (strcasecmp(parameter_name,"end_sphere")==0);
        }
      }
      else if (strcasecmp(parameter_name,"cylinder")==0)
      {
        fSectionParsingDone = false;
        while (!fSectionParsingDone)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          fSectionParsingDone = (strcasecmp(parameter_name,"end_cylinder")==0);
        }
      }
      else if (strcasecmp(parameter_name,"extrusion")==0)
      {
        fSectionParsingDone = false;
        while (!fSectionParsingDone)
        {
          fscanf(airplane_data,"%s %[^\n]",(char *)parameter_name,
                 trailing_comment);

          fSectionParsingDone = (strcasecmp(parameter_name,"end_extrusion")==0);
        }
      }
    }
    fclose(airplane_data);
    {
      // crrc_aero says: Cm = Cm_0 + Cm_a*(Alpha-Alpha_0)
      // So Cm is zero at
      float alpha = ((Cm_a * Alpha_0) - Cm_0 ) / Cm_a;
      // crrc_aero says: CL  = CL_0 + CL_a*(Alpha - Alpha_0);
      float cl = CL_0 + CL_a * (alpha - Alpha_0);

      // sanity check: currently this method does not really work for biplane2.air, as
      // a negative cl is computed. ???
      if (cl < 0.2)
      {
        cl = 0.2;
        printf("airplane: Cm_a=%f   Alpha_0=%f  Cm_0=%f\n", Cm_a, Alpha_0, Cm_0);
        printf("airplane: CL_a=%f               CL_0=%f\n", CL_a,          CL_0);
        printf("airplane: alpha=%f  C_L=%f\n", alpha, cl);
      }

      // v = sqrt(m * g * 2 / (F * CL * rho))
      // I'll use conversions to SI units, just to understand the numbers...
      const float foot = 0.3048;               // m
      const float slug = 14.5939041995;        // kg
      float m    = Mass * slug;                // kg
      float g    = 9.81;                       // m/s^2
      float F    = S_ref * (foot*foot);        // m^2
      float rho  = 1.225;                      // kg/m^3
      float v = sqrt(m * g * 2 / (F * cl * rho));
      //      std::cout << "trimmedFlightVelocity = " << v << " m/s\n";
      trimmedFlightVelocity = v / foot; // ft/s
    }

    std::cout << "CRRC_AirplaneSim_Larcsim::CRRC_AirplaneSim_Larcsim(\"" << filename << "\") done\n";
  }

  if (span_ft == 0.0)
  {
    // no hardpoints --> use reference span
    span_ft = B_ref;
  }
  
  {    
    // try to load new style config and evaluate it    
    std::string name = filename;
    if (name.find(".air") != std::string::npos && name.find(".air") == name.length()-4)
    {
      name = name.substr(0, name.length()-4);
      name += ".xml";
    }
    try
    {
      SimpleXMLTransfer* xml = new SimpleXMLTransfer(name.c_str());
      power = new Power::Power(xml);
      delete xml;
    }
    catch (XMLException e)
    {
      power = (Power::Power*)0;
      std::cout << e.what() << "\n";
    }

    // has power been loaded from new style config?
    if (power == (Power::Power*)0)
    {
      // Do we need a power system because of entry in old style config?
      if (max_thrust > 0)
      {
        SimpleXMLTransfer* xml = new SimpleXMLTransfer();
        
        xml->setAttribute("power.automagic.F", doubleToString(max_thrust*LBF_TO_N));
        xml->setAttribute("power.automagic.V", doubleToString(trimmedFlightVelocity*FT_TO_METERS));
        
        xml->setAttribute("power.automagic.battery.automagic.T", "420");
        
        xml->setAttribute("power.automagic.battery.shaft.J",     "0");
        xml->setAttribute("power.automagic.battery.shaft.brake", "1");
        
        xml->setAttribute("power.automagic.battery.shaft.propeller.D", doubleToString(span*0.2));
        xml->setAttribute("power.automagic.battery.shaft.propeller.H", doubleToString(span*0.2*0.7));
        xml->setAttribute("power.automagic.battery.shaft.propeller.J",      "0");
        xml->setAttribute("power.automagic.battery.shaft.propeller.n_fold", "5");
        
        xml->setAttribute("power.automagic.battery.shaft.engine.automagic.omega_p", "2827");
        xml->setAttribute("power.automagic.battery.shaft.engine.automagic.eta_opt", "0.78");
        xml->setAttribute("power.automagic.battery.shaft.engine.automagic.eta",     "0.7");
                
        power = new Power::Power(xml);
                        
        delete xml;
      }
      else
        power = new Power::Power();
    }
  }
}
/*}}}*/

CRRC_AirplaneSim_002::CRRC_AirplaneSim_002(SimpleXMLTransfer* xml) : CRRC_AirplaneSim("fdm_larcsim.dat") /*{{{*/
{
  SimpleXMLTransfer* i;
  SimpleXMLTransfer* cfg = XMLModelFile::getConfig(xml);
  
  /**
   * Tracks wingspan [m]
   */
  double span = 0;
  span_ft = 0.0;
  
  {    
    double to_ft;
    
    switch (xml->getInt("aero.units"))
    {
     case 0:
      to_ft = 1;
      break;
     case 1:
      to_ft = M_TO_FT;
      break;
     default:
      {
        throw std::runtime_error("Unknown units in aero");
      }
      break;
    }
    i = xml->getChild("aero.ref");
    C_ref = i->getDouble("chord") * to_ft;
    B_ref = i->getDouble("span") * to_ft;
    S_ref = i->getDouble("area") * to_ft * to_ft;
    U_ref = i->getDouble("speed") * to_ft;
  }
  
  i = xml->getChild("aero.misc");
  Alpha_0  = i->getDouble("Alpha_0");
  eta_loc  = i->getDouble("eta_loc");
  CG_arm   = i->getDouble("CG_arm");
  span_eff = i->getDouble("span_eff");
  
  i = xml->getChild("aero.m");
  Cm_0  = i->getDouble("Cm_0");
  Cm_a  = i->getDouble("Cm_a");
  Cm_q  = i->getDouble("Cm_q");
  Cm_de = i->getDouble("Cm_de");
  
  i = xml->getChild("aero.lift");
  CL_0    = i->getDouble("CL_0");
  CL_max  = i->getDouble("CL_max");
  CL_min  = i->getDouble("CL_min");
  CL_a    = i->getDouble("CL_a");
  CL_q    = i->getDouble("CL_q");
  CL_de   = i->getDouble("CL_de");
  CL_drop = i->getDouble("CL_drop");
  CL_CD0  = i->getDouble("CL_CD0");
  
  i = xml->getChild("aero.drag");
  CD_prof  = i->getDouble("CD_prof");
  Uexp_CD  = i->getDouble("Uexp_CD");
  CD_stall = i->getDouble("CD_stall");
  CD_CLsq  = i->getDouble("CD_CLsq");
  CD_AIsq  = i->getDouble("CD_AIsq");
  CD_ELsq  = i->getDouble("CD_ELsq");
  
  i = xml->getChild("aero.Y");
  CY_b  = i->getDouble("CY_b");
  CY_p  = i->getDouble("CY_p");
  CY_r  = i->getDouble("CY_r");
  CY_dr = i->getDouble("CY_dr");
  CY_da = i->getDouble("CY_da");
  
  i = xml->getChild("aero.l");
  Cl_b  = i->getDouble("Cl_b");
  Cl_p  = i->getDouble("Cl_p");
  Cl_r  = i->getDouble("Cl_r");
  Cl_dr = i->getDouble("Cl_dr");
  Cl_da = i->getDouble("Cl_da");
  
  i = xml->getChild("aero.n");
  Cn_b  = i->getDouble("Cn_b");
  Cn_p  = i->getDouble("Cn_p");
  Cn_r  = i->getDouble("Cn_r");
  Cn_dr = i->getDouble("Cn_dr");
  Cn_da = i->getDouble("Cn_da");
  
  
  {
    double to_slug;
    double to_slug_ft_ft;
    
    i = cfg->getChild("mass_inertia");
    switch (i->getInt("units"))
    {
     case 0:    
      to_slug       = 1;
      to_slug_ft_ft = 1;
      break;
     case 1:
      to_slug       = KG_TO_SLUG;
      to_slug_ft_ft = KG_M_M_TO_SLUG_FT_FT;
      break;
     default:
      {
        throw std::runtime_error("Unknown units in mass_inertia");
      }
      break;
    }
    Mass  = i->getDouble("Mass") * to_slug;
    I_xx  = i->getDouble("I_xx") * to_slug_ft_ft;
    I_yy  = i->getDouble("I_yy") * to_slug_ft_ft;
    I_zz  = i->getDouble("I_zz") * to_slug_ft_ft;
    I_xz  = i->getDouble("I_xz") * to_slug_ft_ft;
  }
  
  {
    Wheel              wheel;
    SimpleXMLTransfer* e, f;
    unsigned int       uSize;
    double             x, y, z;
    double             dist;
    double             to_ft;
    double             to_lbf_per_ft;
    double             to_lbf_s_per_ft;
    
    // let's assume that there is nothing below the CG:
    dZLow    = 0;
    
    // let's assume that there is nothing distant from the CG:
    dMaxSize = 0;
    
    wheels.clear();
    
    i = xml->getChild("wheels");
    switch (i->getInt("units"))
    {
     case 0:    
      to_ft           = 1;
      to_lbf_per_ft   = 1;
      to_lbf_s_per_ft = 1;
      break;
     case 1:
      to_ft           = M_TO_FT;
      to_lbf_per_ft   = FT_TO_M / LBF_TO_N;
      to_lbf_s_per_ft = FT_TO_M / LBF_TO_N;
      break;
     default:
      {
        throw std::runtime_error("Unknown units in wheels");
      }
      break;
    }
    uSize = i->getChildCount();
    for (unsigned int n=0; n<uSize; n++)
    {
      e = i->getChildAt(n);
      
      x = e->getDouble("pos.x") * to_ft;
      y = e->getDouble("pos.y") * to_ft;
      z = e->getDouble("pos.z") * to_ft;
      
      wheel.v_P                = CRRCMath::Vector3(x, y, z);
      wheel.spring_constant    = e->getDouble("spring.constant") * to_lbf_per_ft;
      wheel.spring_damping     = e->getDouble("spring.damping")  * to_lbf_s_per_ft;
      wheel.percent_brake      = e->getDouble("percent_brake");
      wheel.caster_angle_rad   = e->getDouble("caster_angle_rad");
      
      if (e->indexOfChild("steering") >= 0)
      {
        std::string s = e->getString("steering.mapping", "NOTHING");
        wheel.steering_max_angle = e->getDouble("steering.max_angle", 1.0);
        
        if (s == "RUDDER")
        {
          wheel.steering_mapping = RUDDER;
        }
        else if (s == "AILERON")
        {
          wheel.steering_mapping = AILERON;
        }
        else if (s == "ELEVATOR")
        {
          wheel.steering_mapping = ELEVATOR;
        }
        else if (s == "NOTHING")
        {
          wheel.steering_mapping = NOTHING;
        }
        else
        {
          wheel.steering_mapping = NOTHING;
          std::cerr << "fdm: unknown steering mapping attribute: \"" << s << "\"" << std::endl;
          std::cerr << "fdm: falling back to NOTHING" << std::endl;
        }
      }
      else
      {
        wheel.steering_mapping   = NOTHING;
        wheel.steering_max_angle = 0;
      }
      wheels.push_back(wheel);
      
      // track wingspan
      if (span < y)
        span = y;
      // lowest point?
      if (dZLow < z)
        dZLow = z;
      // far away (Z distance is assumed to be low)?
      dist = x*x + y*y;
      if (dist > dMaxSize)
        dMaxSize = dist;
    }
    dMaxSize = sqrt(dMaxSize);
    span_ft  = 2 * span;
    span     = span_ft * FT_TO_METERS;
    
    // just in case: if there were no hardpoints, use the reference span
    if (span_ft == 0.0)
    {
      span_ft = B_ref;
    }
  }
  
    
  // calculate velocity in trimmed flight
  {
    // crrc_aero says: Cm = Cm_0 + Cm_a*(Alpha-Alpha_0)
    // So Cm is zero at
    float alpha = ((Cm_a * Alpha_0) - Cm_0 ) / Cm_a;
    // crrc_aero says: CL  = CL_0 + CL_a*(Alpha - Alpha_0);
    float cl = CL_0 + CL_a * (alpha - Alpha_0);
    
    // sanity check: currently this method does not really work for biplane2.air, as
    // a negative cl is computed. ???
    if (cl < 0.2)
    {
      cl = 0.2;
      printf("airplane: Cm_a=%f   Alpha_0=%f  Cm_0=%f\n", Cm_a, Alpha_0, Cm_0);
      printf("airplane: CL_a=%f               CL_0=%f\n", CL_a,          CL_0);
      printf("airplane: alpha=%f  C_L=%f\n", alpha, cl);
    }
    
    // v = sqrt(m * g * 2 / (F * CL * rho))
    // I'll use conversions to SI units, just to understand the numbers...
    const float foot = 0.3048;               // m
    const float slug = 14.5939041995;        // kg
    float m    = Mass * slug;                // kg
    float g    = 9.81;                       // m/s^2
    float F    = S_ref * (foot*foot);        // m^2
    float rho  = 1.225;                      // kg/m^3
    float v = sqrt(m * g * 2 / (F * cl * rho));
    std::cout << "trimmedFlightVelocity = " << v << " m/s\n";
    trimmedFlightVelocity = v / foot; // ft/s
  }
  
  // possibly there is no power description:
  if (cfg->indexOfChild("power") < 0)
  {
    power = new Power::Power();
  }
  else
  {
    power = new Power::Power(cfg);
  }  
}
/*}}}*/

CRRC_AirplaneSim_002::~CRRC_AirplaneSim_002()
{
  delete power;
}

double CRRC_AirplaneSim_002::getPhi()
{
  return(eom.conv.euler.r[0]);
}

double CRRC_AirplaneSim_002::getTheta()
{
  return(eom.conv.euler.r[1]);
}

double CRRC_AirplaneSim_002::getPsi()
{
  return(eom.conv.euler.r[2]);
}

CRRCMath::Vector3 CRRC_AirplaneSim_002::getPos()
{
  return(eom.pos.val);
}


/**
 * 
 * jwtodo
 *  gut:
 *   skimmer
 *   arkjan
 *   apogee
 *   allegro
 *   mav
 *   delta*
 *   zip60
 *   sport
 *   viele andere
 * 
 *  schlecht:
 *   superzagi
 *   quickie
 *   zagi
 *   zagi-xs
 * 
 */
/** \brief Calculate influence of gear and hardpoints
 *
 *  This method calculates the forces and moments on
 *  the airplane's body caused by the gear or other
 *  hardpoints touching the ground.
 *
 *  \param inputs   Current control inputs
 */
void CRRC_AirplaneSim_002::gear(TSimInputs* inputs)      /*{{{*/
{
  /*
   * Constants & coefficients for tyres on tarmac - ref [1]
   */

  /* skid function looks like:
   *
   *           mu  ^
   *               |
   *       max_mu  |       +
   *               |      /|
   *   sliding_mu  |     / +------
   *               |    /
   *               |   /
   *               +--+------------------------>
   *               |  |    |      sideward V
   *               0 bkout skid
   *                V     V
   */

  const DATA sliding_mu   = 0.55;
  const DATA rolling_mu   = 0.01;
  const DATA max_brake_mu = 0.6;
  const DATA max_mu       = 0.8;
  const DATA bkout_v      = 0.1;
  const DATA skid_v       = 1.0;

  /*
   * Local data variables
   */

  DATA reaction_normal_force;   /* wheel normal (to rwy) force */
  DATA cos_wheel_hdg_angle, sin_wheel_hdg_angle;  /* temp storage */
  DATA steering_angle_rad;
  DATA v_wheel_forward, v_wheel_sideward,  abs_v_wheel_sideward;
  DATA forward_mu, sideward_mu; /* friction coefficients */
  DATA beta_mu;                 /* breakout friction slope */
  DATA forward_wheel_force, sideward_wheel_force;
  DATA z_earth;

  int i;                        /* per wheel loop counter */
  int num_wheels = wheels.size();

  CRRCMath::Vector3 temp3a, temp3b, tempF, tempM;
  CRRCMath::Vector3 v_V_wheel_local;
  CRRCMath::Vector3 v_F_wheel_local;
  
  /* wheel offset from cg,  X-Y-Z */
  CRRCMath::Vector3 v_P_wheel_cg_body;
  
  /* wheel offset from cg,  N-E-D */
  CRRCMath::Vector3 v_P_wheel_cg_local;
  
  /* wheel offset from rwy, N-E-D */
  CRRCMath::Vector3 v_P_wheel_rwy_local;
  
  /*
   * Execution starts here
   */

  beta_mu = max_mu/(skid_v-bkout_v);
  
  v_F_gear = CRRCMath::Vector3();  /* Initialize sum of forces... */
  v_M_gear = CRRCMath::Vector3();  /* ...and moments  */

  
  for (i=0;i<num_wheels;i++)     /* Loop for each wheel */
  {
    /*========================================*/
    /* Calculate wheel position w.r.t. runway */
    /*========================================*/

    /* First calculate wheel location w.r.t. cg in body (X-Y-Z) axes... */

    v_P_wheel_cg_body = wheels[i].v_P;

    /* then converting to local (North-East-Down) axes... */

    v_P_wheel_cg_local = eom.conv.local(v_P_wheel_cg_body);    
    
    /* Add wheel offset to cg location in local axes */

    v_P_wheel_rwy_local = v_P_wheel_cg_local + eom.pos.val;

    /*============================*/
    /* Calculate wheel velocities */
    /*============================*/

    /* contribution due to angular rates */

    temp3a = eom.angvel.val * v_P_wheel_cg_body;

    /* transform into local axes */

    temp3b = eom.conv.local(temp3a);
    
    /* plus contribution due to cg velocities */

    v_V_wheel_local = temp3b + eom.conv.local(eom.vel.val);

    /*===========================================*/
    /* Calculate forces & moments for this wheel */
    /*===========================================*/

    /* Steering angle */
    /* First, determine the control input for this wheel */
    switch(wheels[i].steering_mapping)
    {
      case AILERON:
        steering_angle_rad = inputs->aileron;
        break;
      
      case RUDDER:
        steering_angle_rad = inputs->rudder;
        break;
      
      case ELEVATOR:
        steering_angle_rad = inputs->elevator;
        break;

      case NOTHING:
      default:
        steering_angle_rad = 0.0;
        break;
    }
    /* Then calculate the real angle. Full control input (+-0.5)
       shall result in full wheel deflection  */
    steering_angle_rad = 2 * steering_angle_rad * wheels[i].steering_max_angle;

    /* Calculate sideward and forward velocities of the wheel
     in the runway plane     */

    double Psi = eom.conv.euler.r[2];
    
    DATA tmp_angle = wheels[i].caster_angle_rad + steering_angle_rad + Psi;
    cos_wheel_hdg_angle = cos(tmp_angle);
    sin_wheel_hdg_angle = sin(tmp_angle);

    v_wheel_forward  = v_V_wheel_local.r[0]*cos_wheel_hdg_angle
                     + v_V_wheel_local.r[1]*sin_wheel_hdg_angle;
    v_wheel_sideward = v_V_wheel_local.r[1]*cos_wheel_hdg_angle
                     - v_V_wheel_local.r[0]*sin_wheel_hdg_angle;    

    /* Calculate normal load force (simple spring constant) */

    reaction_normal_force = 0.;

    z_earth = -1 * Global::scenery->getHeight(v_P_wheel_rwy_local.r[0], v_P_wheel_rwy_local.r[1]);

    if( v_P_wheel_rwy_local.r[2] > z_earth)
    {
      // Forces are in lbf here, lengths in ft, velocities in ft/s. 
      // So:
      // 1 slug * 1 ft / s^2 = spring_constant * 1 ft - 1 ft/s  * spring_damping
      // spring_constant  = slug / s^2 = lbf / ft
      // spring_damping   = slug / s   = lbf * s / ft
      reaction_normal_force = wheels[i].spring_constant*(z_earth-v_P_wheel_rwy_local.r[2])
                             - v_V_wheel_local.r[2]*wheels[i].spring_damping;
    }

    /* Calculate friction coefficients */

    forward_mu = (max_brake_mu - rolling_mu)*wheels[i].percent_brake + rolling_mu;
    abs_v_wheel_sideward = sqrt(v_wheel_sideward*v_wheel_sideward);
    sideward_mu = sliding_mu;
    if (abs_v_wheel_sideward < skid_v)
      sideward_mu = (abs_v_wheel_sideward - bkout_v)*beta_mu;
    if (abs_v_wheel_sideward < bkout_v)
      sideward_mu = 0.;

    /* Calculate foreward and sideward reaction forces */

    forward_wheel_force  =   forward_mu*reaction_normal_force;
    sideward_wheel_force =  sideward_mu*reaction_normal_force;
    if(v_wheel_forward < 0.)
      forward_wheel_force = -forward_wheel_force;
    if(v_wheel_sideward < 0.)
      sideward_wheel_force = -sideward_wheel_force;

    /* Rotate into local (N-E-D) axes */

    v_F_wheel_local.r[0] = forward_wheel_force *cos_wheel_hdg_angle
                         - sideward_wheel_force*sin_wheel_hdg_angle;
    v_F_wheel_local.r[1] = forward_wheel_force *sin_wheel_hdg_angle
                         + sideward_wheel_force*cos_wheel_hdg_angle;
    v_F_wheel_local.r[2] = reaction_normal_force;

    /* Convert reaction force from local (N-E-D) axes to body (X-Y-Z) */

    tempF = eom.conv.body(v_F_wheel_local);

    /* Calculate moments from force and offsets in body axes */

    tempM = v_P_wheel_cg_body * tempF;

    /* Sum forces and moments across all wheels */

    v_F_gear += tempF;
    v_M_gear += tempM;
  }
}
/*}}}*/

/**
        X       Aerodynamic force, lbs, in X-axis (+ forward)
        Y       Aerodynamic force, lbs, in Y-axis (+ right)
        Z       Aerodynamic force, lbs, in Z-axis (+ down)
        L       Aero. moment about X-axis (+ roll right), ft-lbs
        M       Aero. moment about Y-axis (+ pitch up), ft-lbs
        N       Aero. moment about Z-axis (+ nose right), ft-lbs

        0       Subscript implying initial, or nominal, value
        u       X-axis component of airspeed (ft/sec) (+ forward)
        v       Y-axis component of airspeed (ft/sec) (+ right)
        w       Z-axis component of airspeed (ft/sec) (+ down)
        p       X-axis ang. rate (rad/sec) (+ roll right), rad/sec
        q       Y-axis ang. rate (rad/sec) (+ pitch up), rad/sec
        r       Z-axis ang. rate (rad/sec) (+ yaw right), rad/sec
        beta    Angle of sideslip, degrees (+ wind in RIGHT ear)
        da      Aileron deflection, degrees (+ left ail. TE down)
        de      Elevator deflection, degrees (+ trailing edge down)
        dr      Rudder deflection, degrees (+ trailing edge LEFT)
 */
void CRRC_AirplaneSim_002::aero(SCALAR dt, TSimInputs* inputs) /*{{{*/
  // Calculate forces and moments for the current time step.  If Initialize is
  // zero, then re-initialize coefficients by reading in the coefficient file.
{
  SCALAR elevator, aileron, rudder;

  SCALAR Phat, Qhat, Rhat;  // dimensionless rotation rates
  SCALAR CL_left, CL_cent, CL_right; // CL values across the span
  SCALAR dCL_left,dCL_cent,dCL_right; // stall-induced CL changes
  SCALAR dCD_left,dCD_cent,dCD_right; // stall-induced CD changes
  SCALAR dCl,dCn,dCl_stall,dCn_stall;  // stall-induced roll,yaw moments
  SCALAR dCm_stall;  // Stall induced pitching moment.
  SCALAR CL_wing, CD_all, CD_scaled, Cl_w;
  SCALAR Cl_r_mod,Cn_p_mod;
  SCALAR CL,CD,Cl,Cm,Cn;
  SCALAR QS;
  
  CRRCMath::Vector3 C_xyz;

  CRRCMath::Vector3 v_R_omega_atmo;
  
  CRRCMath::Matrix33 G;
  CRRCMath::Matrix33 m_V_body;
  
  double V_rel_wind;
  double Alpha;
  double Beta;
    
  elevator = inputs->elevator;
  aileron  = inputs->aileron;
  rudder   = inputs->rudder;
  
  CRRCMath::Vector3 V_airmass_body  = eom.conv.body(v_V_local_airmass);

  // velocity of body relative to airmass
  v_V_body = eom.vel.val - V_airmass_body;

  V_rel_wind = v_V_body.length();

  /* Calculate alpha and beta */

  if (v_V_body.r[0] == 0)
    Alpha = 0;
  else
    Alpha = atan2( v_V_body.r[2], v_V_body.r[0] );

  if (V_rel_wind == 0)
    Beta = 0;
  else
    Beta = asin( v_V_body.r[1]/ V_rel_wind );

  
//  std::cout << Alpha << " ";
  
  // ---------------
  //
  double Sin_alpha = sin(Alpha);
  double Cos_alpha = cos(Alpha);
  double Cos_beta  = cos(Beta);

  /* compute gradients of Local velocities w.r.t. Body coordinates
      G_11  =  dU_local/dx_body
      G_12  =  dU_local/dy_body   etc.  */

  G = m_V_atmo_rwy * eom.conv.mat.trans();

  /* now compute gradients of Body velocities w.r.t. Body coordinates */
  /*  U_body_x  =  dU_body/dx_body   etc.  */

  m_V_body = eom.conv.mat * G;
  
  /* set rotation rates of airmass motion */
  v_R_omega_atmo.r[0] =  m_V_body.v[2][0];
  v_R_omega_atmo.r[1] = -m_V_body.v[2][1];
  v_R_omega_atmo.r[2] =  m_V_body.v[1][2];

  if (V_rel_wind != 0)
  {
    /* set net effective dimensionless rotation rates */
    Phat = (eom.angvel.val.r[0] - v_R_omega_atmo.r[0]) * B_ref / (2.0*V_rel_wind);
    Qhat = (eom.angvel.val.r[1] - v_R_omega_atmo.r[1]) * C_ref / (2.0*V_rel_wind);
    Rhat = (eom.angvel.val.r[2] - v_R_omega_atmo.r[2]) * B_ref / (2.0*V_rel_wind);
  }
  else
  {
    Phat=0;
    Qhat=0;
    Rhat=0;
  }

  /* compute local CL at three spanwise locations */
  CL_left  = CL_0 + CL_a*(Alpha - Alpha_0 - Phat*eta_loc);
  CL_cent  = CL_0 + CL_a*(Alpha - Alpha_0               );
  CL_right = CL_0 + CL_a*(Alpha - Alpha_0 + Phat*eta_loc);

  /* set CL-limit changes */
  dCL_left  = 0.;
  dCL_cent  = 0.;
  dCL_right = 0.;

  stalling=0;
  if (CL_left  > CL_max)
  {
    dCL_left  = CL_max-CL_left -CL_drop;
    stalling=1;
  }

  if (CL_cent  > CL_max)
  {
    dCL_cent  = CL_max-CL_cent -CL_drop;
    stalling=1;
  }

  if (CL_right > CL_max)
  {
    dCL_right = CL_max-CL_right -CL_drop;
    stalling=1;
  }

  if (CL_left  < CL_min)
  {
    dCL_left  = CL_min-CL_left -CL_drop;
    stalling=1;
  }

  if (CL_cent  < CL_min)
  {
    dCL_cent  = CL_min-CL_cent -CL_drop;
    stalling=1;
  }

  if (CL_right < CL_min)
  {
    dCL_right = CL_min-CL_right -CL_drop;
    stalling=1;
  }
  
//  std::cout << CL_left << " " << CL_cent << " " << CL_right << " ";

  /* set average wing CL */
  CL_wing = CL_0 + CL_a*(Alpha-Alpha_0)
    + 0.25*dCL_left + 0.5*dCL_cent + 0.25*dCL_right;

  /* correct profile CD for CL dependence and aileron dependence */
  CD_all = CD_prof
    + CD_CLsq *(CL_wing-CL_CD0)*(CL_wing-CL_CD0)
    + CD_AIsq * aileron        * aileron
    + CD_ELsq * elevator       * elevator;

  /* scale profile CD with Reynolds number via simple power law */
  if (V_rel_wind > 0.1)
  {
    CD_scaled = CD_all*pow(((double)V_rel_wind/(double)U_ref),Uexp_CD);
  }
  else
  {
    CD_scaled=CD_all;
  }

//  std::cout << CL_wing << " " << CD_scaled << " ";
  
  /* Scale lateral cross-coupling derivatives with wing CL */
  Cl_r_mod = Cl_r*CL_wing/CL_0;
  Cn_p_mod = Cn_p*CL_wing/CL_0;

  //std::cout << Cl_r_mod << " " << Cn_p_mod << " ";
  
  /* total average CD with induced and stall contributions */
  dCD_left  = CD_stall*dCL_left *dCL_left ;
  dCD_cent  = CD_stall*dCL_cent *dCL_cent ;
  dCD_right = CD_stall*dCL_right*dCL_right;

  //std::cout << dCD_left << " " << dCD_cent << " " << dCD_right << " ";
  
  /* total CL, with pitch rate and elevator contributions */
  CL = (CL_wing + CL_q*Qhat + CL_de*elevator)*Cos_alpha;

  //std::cout << CL << " ";
  
  /* assymetric stall will cause roll and yaw moments */
  dCl =  0.45*-1*(dCL_right-dCL_left)*0.5*eta_loc;
  dCn =  0.45*(dCD_right-dCD_left)*0.5*eta_loc;
  dCm_stall = (0.25*dCL_left + 0.5*dCL_cent + 0.25*dCL_right)*CG_arm;

  //std::cout << dCl << " " << dCn << " " << dCm_stall << " ";
  
  /* stall-caused moments in body axes */
  dCl_stall = dCl*Cos_alpha - dCn*Sin_alpha;
  dCn_stall = dCl*Sin_alpha + dCn*Cos_alpha;

  //std::cout << dCl_stall << " " << dCn_stall << " ";
  
  /* total CD, with induced and stall contributions */

  Cl_w = Cl_b*Beta  + Cl_p*Phat + Cl_r_mod*Rhat
    + dCl_stall  + Cl_da*aileron;
  CD = CD_scaled
    + (CL*CL + 32.0*Cl_w*Cl_w)*S_ref/(B_ref*B_ref*3.14159*span_eff)
      + 0.25*dCD_left + 0.5*dCD_cent + 0.25*dCD_right;

  //std::cout << Cl_w << " " << CD << " ";
  
  /* total forces in body axes */
  C_xyz.r[0] = -CD*Cos_alpha + CL*Sin_alpha*Cos_beta*Cos_beta;
  C_xyz.r[2] = -CD*Sin_alpha - CL*Cos_alpha*Cos_beta*Cos_beta;
  C_xyz.r[1] = CY_b*Beta  + CY_p*Phat + CY_r*Rhat + CY_dr*rudder;

  //std::cout << Cx << " " << Cz << " " << Cy << " ";

  /* total moments in body axes */
  Cl =        Cl_b*Beta  + Cl_p*Phat + Cl_r_mod*Rhat + Cl_dr*rudder
    + dCl_stall                               + Cl_da*aileron;
  Cn =        Cn_b*Beta  + Cn_p_mod*Phat + Cn_r*Rhat + Cn_dr*rudder
    + dCn_stall                               + Cn_da*aileron;
  Cm = Cm_0 + Cm_a*(Alpha-Alpha_0) +dCm_stall
    + Cm_q*Qhat                 + Cm_de*elevator;

  /* set dimensional forces and moments */
  QS = 0.5*DENSITY*V_rel_wind*V_rel_wind * S_ref;

//  std::cout << ">" << Cm_0 << " " << Cm_a << " " << Alpha_0 << " " << dCm_stall << " " << Cm_q << " " << Cm << "< ";
  
  // std::cout << Cl << " " << Cm << " " << Cn << " ";
  
//  std::cout << DENSITY << "density " << QS << "QS ";

  v_F_aero = C_xyz * QS;

  v_M_aero = CRRCMath::Vector3(Cl * QS * B_ref, Cm * QS * C_ref, Cn * QS * B_ref); 
  
  /*
  std::cout << (Cx * QS) << " " << (Cy * QS) << " " << (Cz * QS) << " " 
            << (Cl * QS * B_ref) << " " << (Cm * QS * C_ref) << " " << (Cn * QS * B_ref) << "\n";
   */



//  std::cout << "\n";
  
#if (EOM_TEST == 1)
  {
    double ele = 1.8*(-inputs->elevator-eom.angvel.val.r[1]);
    double ail = 1.8*(inputs->aileron-eom.angvel.val.r[0]);
    double rud = 1.8*(0.5*inputs->aileron-eom.angvel.val.r[2]);
   
    v_F_aero = CRRCMath::Vector3(4*inputs->throttle-0.8*eom.vel.val.r[0], -2.8*eom.vel.val.r[1], -2.8*eom.vel.val.r[2]);
    v_M_aero = CRRCMath::Vector3(ail, ele, rud);
    stalling = 0;
  }
#endif  
    
//  eom.pos.val.print("", "\n");
  
  
//  std::cout << "|V|=" << V_rel_wind << ", ";
//  std::cout << "alpha=" << Alpha << ", ";
//  std::cout << "beta=" << Beta << ", ";
//  V_body.print("V=", ", ");
//  v_F_aero.print("f=", ", ");
//  v_M_aero.print("m=", "\n");
}
/*}}}*/

void CRRC_AirplaneSim_002::aero_init() /*{{{*/
{
  stalling = 0;
  v_F_aero = CRRCMath::Vector3();
  v_M_aero = CRRCMath::Vector3();
}
/*}}}*/

double CRRC_AirplaneSim_002::getPropFreq()
{
  return(power->getPropFreq());
};
